/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*---------------------------------------------------------------------------
 *		
 * MODULE NAME:	crashMessage
 *
 * DESCRIPTION:	These routines are used to print out diagnostic information 
 *		when an RCI application program dies as a result of a 
 *		hardware execption at the control level. "Hardware
 *		exception" means pretty much anything that would
 *		cause a regular UNIX program to core dump.
 *
 * AUTHOR/DATE:	John Lloyd, Fall 1987
 *
 * MODIFICATION HISTORY:
 *
 *		John Lloyd, Fall 1988: Added support for SUN4/SPARC.
 *
 *--------------------------------------------------------------------------*/

#ifndef SIMONLY_CFG
#if defined(Sun4_TARGET) || defined(Vxw_TARGET) || defined(UVax_TARGET)

#ifndef Vxw_TARGET
#include <stdio.h>
#else
#include <vxWorks.h>
#include <stdioLib.h>
#endif

#include "rciConfig.h"
#include "crashinfo.h" /* crash information structure */

#if defined(SUN4_CFG) || defined(Vxw_TARGET)
crashMessage (s, cp, onunix)
char *s;
register CRASHINFO *cp;
int onunix;
{
	/* This routine analyses the crash information provided in the
	   indicated CRASHINFO structure and provides a formatted
	   message indiacting the nature of the crash (i.e., 'integer
	   divide by zero', or 'access violation'), as well as useful
	   information like the program counter where the crash occurred,
	   the address in the case of memory access faults, etc.

	   The crash information itself is returned by the ka driver (for
	   task running on the auxiliary processors), or by an RCI
	   device driver (for task executing on the arbiter).
	*/
	
	switch (cp->type)
	 { default:
	    { sprintf (s,
	      "ERROR -- bogus crash type specified (%d)", cp->type);
	      break;
	    }
	   case CI_NULL:
	    { sprintf (s,
	      "no exit type specified");
	      break;
	    }
	   case CI_ARITHTRAP:	/* kill with SIGFPE */
	    { 
	      char *subtype = 0;

	      switch (cp->un.subtype)
	       {
		 case (CI_AT_IDIVZ): /* integer divide-by-zero */
		  { subtype = "integer divide by zero trap";
		    break;
		  }
		 case (CI_AT_FINEXACT): /* floating overflow */
		  { subtype = "floating inexact trap";
		    break;
		  }
		 case (CI_AT_FDIVZ): /* floating divide-by-zero */
		  { subtype = "floating divide by zero trap";
		    break;
		  }
		 case (CI_AT_FUNDERF): /* floating underflow */
		  { subtype = "floating underflow trap";
		    break;
		  }
		 case (CI_AT_FOPERAND): /* floating operand */
		  { subtype = "bad floating operand trap";
		    break;
		  }
		 case (CI_AT_FOVERF): /* floating overflow */
		  { subtype = "floating overflow trap";
		    break;
		  }
		 case (CI_AT_BOUNDS): /* operand out of bounds */
		  { subtype = "out of bounds trap";
		    break;
		  }
		 case (CI_AT_OVERFTRAP): /* overflow trap */
		  { subtype = "overflow trap";
		    break;
		  }
		 case (CI_AT_FNAN): /* not a number */
		  { subtype = "not-a-number trap";
		    break;
		  }
		 case (CI_AT_FDISABLED): /* FPU disabled */
		  { subtype = "FPU disabled trap";
		    break;
		  }
		 case (CI_AT_FERROR): /* floating point error */
		  { subtype = "general FPU error trap";
		    break;
		  }
	       }
	      if (subtype)
	       { sprintf (s, "arithmetic trap, pc = 0x%x (%s)", 
			  cp->pc, subtype);
	       }
	      else
	       { sprintf (s, "arithmetic trap, pc = 0x%x (%s %d)", 
			  cp->pc, "ERROR -- bad subtype", cp->un.subtype);
	       }
	      break;
	    }
	   case CI_ILLINST:	/* UNIX: send a SIGILL */
	    { 
	      char *subtype = 0;

	      switch (cp->un.subtype)
	       { 
		 case (CI_II_UNIMP):
		  { subtype = "unimplemented instruction";
		    break;
		  }
		 case (CI_II_PRIV):
		  { subtype = "privileged instruction";
		    break;
		  }
		 case (CI_II_SYSCALL):
		  { subtype = "syscall instruction";
		    break;
		  }
		 case (CI_II_COPROT):
		  { subtype = "coprocessor prototcol error";
		    break;
		  }
	       }
	      if (subtype)
	       { sprintf (s, "illegal instruction, pc 0x%x (%s)", 
			  cp->pc, subtype);
	       }
	      else
	       { sprintf (s, "illegal instruction, pc 0x%x (%s %d)", 
			  cp->pc, "ERROR -- bad subtype", cp->un.subtype);
	       }
	      break;
	    }
	   case CI_TRACE:
	    { sprintf (s, "trace trap (!?) -- better call John or Mike!");
	      break;
	    }
	   case CI_MEMFAULT:
	    { 
	      int faultbits = cp->un.memfault.be;
	      int addr = cp->un.memfault.addr;

	      sprintf (s,
		 "memory violation, pc 0x%x, address 0x%x, code 0x%x\n",
		       cp->pc, addr, faultbits);
	      break;
	    }
	   case CI_FRAMEFAULT:
	    { sprintf (s, "stack frame fault, pc 0x%x\n", cp->pc);
	      break;
	    }
	 }
}

#define LINEWIDTH	80

crashStackDump (cp)
CRASHINFO *cp;
{
	int k, j;

	if (cp->type == CI_FRAMEFAULT)
	 { fprintf (stderr, "\nNo stack trace available\n");
	   return;
	 }
	fprintf (stderr, "\nStack trace follows --\n");
	fprintf (stderr, "\n PC values  arguments\n\n");
	for (k=0; k<KAC_MAXCALLS; k++)
	 { CRASHFRAME *frame = &cp->callstack[k];
	   if (!(frame->flags & CI_CF_VALID) ||
		(frame->pc & 0x80000000))
	    { break;
	    }
	   fprintf (stderr, " %8x : ", frame->pc);
	   for (j=0; j<KAC_NARGS; j++)
	    { fprintf (stderr, " %8x", frame->args[j]);
	    }
	   putc ('\n', stderr);
	 }
}

#endif

#if defined(UVax_TARGET)
#include "kacrash.h" /* crash information structure */

crashMessage (s, cp, onunix)
char *s;
register CRASHINFO *cp;
int onunix;
{
	int k, j;

	/* This routine analyses the crash information provided in the
	   indicated CRASHINFO structure and provides a formatted
	   message indiacting the nature of the crash (i.e., 'integer
	   divide by zero', or 'access violation'), as well as useful
	   information like the program counter where the crash occurred,
	   the address in the case of memory access faults, etc.

	   The crash information itself is returned by the ka driver (for
	   task running on the auxiliary processors), or by an RCI
	   device driver (for task executing on the arbiter).
	*/
	
	switch (cp->type)
	 { default:
	    { sprintf (s,
	      "ERROR -- bogus crash type specified (%d)", cp->type);
	      break;
	    }
	   case CI_NULL:
	    { sprintf (s,
	      "no exit type specified");
	      break;
	    }
	   case CI_EXIT:
	    { if (onunix)
	       { sprintf (s,
		 "ERROR -- crash type EXIT shouldn't be set on UNIX");
	       }
	      else
	       { sprintf (s,
		 "program exited with code %d", cp->u.subtype);
		 break;
	       }
	    }
	   case CI_ARITHTRAP:	/* kill with SIGFPE */
	    { char *subtype = 0;

	      switch (cp->u.subtype)
	       {
		 case (CI_AT_IOVERF): /* integer overflow */
		  { subtype = "integer overflow trap";
		    break;
		  }
		 case (CI_AT_IDIVZ): /* integer divide-by-zero */
		  { subtype = "integer divide by zero trap";
		    break;
		  }
		 case (CI_AT_FOVERF): /* floating overflow */
		  { subtype = "floating overflow trap";
		    break;
		  }
		 case (CI_AT_FDIVZ): /* floating divide-by-zero */
		  { subtype = "floating divide by zero trap";
		    break;
		  }
		 case (CI_AT_FUNDERF): /* floating underflow */
		  { subtype = "floating underflow trap";
		    break;
		  }
		 case (CI_AT_DOVERF): /* decimal overflow */
		  { subtype = "decimal overflow trap";
		    break;
		  }
		 case (CI_AT_SUBRANGE): /* subscript range */
		  { subtype = "subscript out of range trap";
		    break;
		  }
		 case (CI_AF_FOVERF): /* floating overflow */
		  { subtype = "floating overflow fault";
		    break;
		  }
		 case (CI_AF_FDIVZ): /* floating divide-by-zero */
		  { subtype = "floating divide by zero fault";
		    break;
		  }
		 case (CI_AF_FUNDERF):
		  { subtype = "floating underflow fault";
		    break;
		  }
	       }
	      if (subtype)
	       { sprintf (s, "arithmetic trap, pc = 0x%x (%s)", 
			  cp->pc, subtype);
	       }
	      else
	       { sprintf (s, "arithmetic trap, pc = 0x%x (%s %d)", 
			  cp->pc, "ERROR -- bad subtype", cp->u.subtype);
	       }
	      break;
	    }
	   case CI_ILLINST:	/* UNIX: send a SIGILL */
	    { char *subtype = 0;

	      switch (cp->u.subtype)
	       { 
		 case (CI_II_RADRMOD):
		  { subtype = "reserved addressing mode";
		    break;
		  }
		 case (CI_II_PRIVINST):
		  { subtype = "privileged instruction";
		    break;
		  }
		 case (CI_II_RESOPRND):
		  { subtype = "reserved operand";
		    break;
		  }
		 case (CI_II_BPT):
		  { subtype = "breakpoint trap"; /* UNIX makes SIGTRAP */
		    break;
		  }
	       }
	      if (subtype)
	       { sprintf (s, "illegal instruction, pc 0x%x (%s)", 
			  cp->pc, subtype);
	       }
	      else
	       { sprintf (s, "illegal instruction, pc 0x%x (%s %d)", 
			  cp->pc, "ERROR -- bad subtype", cp->u.subtype);
	       }
	      break;
	    }
	   case CI_TRACE:
	    { sprintf (s, "trace trap (!?) -- better call John or Mike!");
	      break;
	    }
	   case CI_MEMFAULT:
	    { int faultbits = cp->u.memfault.bits;
	      int addr = cp->u.memfault.vaddr;

	      if (onunix)
	       { switch (faultbits)
		  { case (CI_MF_LENGTH | CI_MF_PAGE | CI_MF_READ | CI_MF_MEM):
		     { sprintf (s, 	/* non-existant virt space */
	"segmentation violation, addr = 0x%x pc = 0x%x", addr, cp->pc);
		       break;
		     }
		    case (CI_MF_PROT | CI_MF_PAGE | CI_MF_READ | CI_MF_MEM):
		     { sprintf (s,
	"bus error or page fault, addr = 0x%x pc = 0x%x", addr, cp->pc);
		       break;
		     }
		    case (CI_MF_PROT | CI_MF_PTE | CI_MF_READ | CI_MF_MEM):
		     { sprintf (s,
	"PANIC -- page table page fault! addr = 0x%x pc = 0x%x",addr, cp->pc);
		       break;
		     }
		    case (CI_MF_PROT | CI_MF_PAGE | CI_MF_READ | CI_MF_QBUS):
		     { sprintf (s,
	"Qbus access fault on read, addr = 0x%x pc = 0x%x\n", addr, cp->pc);
		       break;
		     }
		    case (CI_MF_PROT | CI_MF_PAGE | CI_MF_WRITE | CI_MF_QBUS):
		     { sprintf (s, 
	"Qbus access fault on write, addr = 0x%x pc = 0x%x\n", addr, cp->pc);
		       break;
		     }
		    default:
		     { sprintf (s,
		       "Strange memory fault (0x%x), addr = 0x%x pc = 0x%x",
		       addr, faultbits, cp->pc);
		       break;
		     }
		  }
	       }
	      else
	       { char *where, *what, *type, *access;

		 if ((faultbits & CI_MF_WHERE) == CI_MF_QBUS)
		  { where = "Qbus";
		  }
		 else
		  { where = "memory";
		  }
		 if ((faultbits & CI_MF_WHAT) == CI_MF_PTE)
		  { what = "pte";
		  }
		 else
		  { what = "page";
		  }
		 if ((faultbits & CI_MF_TYPE) == CI_MF_PROT)
		  { type = "protection";
		  }
		 else
		  { type = "length";
		  }
		 if ((faultbits & CI_MF_ACCESS) == CI_MF_READ)
		  { access = "read";
		  }
		 else
		  { access = "write";
		  }
		 sprintf (s,"%s %s %s violation on %s, addr = 0x%x pc = 0x%x",
			  where, what, type, access, addr, cp->pc);
	       }
	    }
	 }
}

#define LINEWIDTH	80

crashStackDump (cp)
CRASHINFO *cp;
{
	int k, j;

	fprintf (stderr, "\nStack trace follows --\n");
	fprintf (stderr, "\n PC values  arguments\n\n");
	for (k=0; k<KAC_MAXCALLS; k++)
	 { CRASHFRAME *frame = &cp->callstack[k];
	   if (!(frame->flags & CI_CF_VALID) ||
		(frame->pc & 0x80000000))
	    { break;
	    }
	   fprintf (stderr, " %8x : ", frame->pc);
	   if ((frame->flags & CI_CF_ARGLOST) ||
	       (frame->nargs == 255) )
	    { fprintf (stderr, "(args not available)\n");
	    }
	   else
	    { int wc = 0;
	      int wordsPerLine = (LINEWIDTH-12)/9;
	      int newline = 0;

	      for (j=0; j<frame->nargs; j++)
	       { fprintf (stderr, " %8x", cp->callargs[frame->argindex+j]);
		 if (++wc == wordsPerLine)
		  { putc ('\n', stderr);
		    wc = 0;
		    newline = 1;
		  }
		 else
		  { newline = 0;
		  }
	       }
	      if (!newline)
	       { putc ('\n', stderr);
	       }
	    }
	 }
}

#endif

#else /* !defined(Sun4_TARGET,Vxw_TARGET,UVax_TARGET) */
	 
crashMessage()
{
}

crashStackDump()
{
}

#endif
#else /* defined(SIMONLY_CFG) */

crashMessage()
{
	 
}
  
crashStackDump()
{
	 
}
#endif
