/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "../h/stringDefs.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../h/robot_jls.h"
#include "../h/robotMath.h"
#include "../h/utilFxns.h"

extern char *rciGetenv();

static FILE *stream;
static char *textp;
static char textBuf[256] = {'\0'};
static int lineNo = 0;
static int atEOF = 0;

/* ----- Error Messages. Defined externally to save space ----- */

static char *rbtCantGetJLSMsg	= "%s -- rbt %s: can't get JLS structure\n";
static char *noEnvMsg		= "%s -- environment variable %s not set\n";
static char *cantOpenFileMsg	= "%s -- can't open file %s\n";
static char *expectedItemMsg    = "%s -- file %s: %s expected, line %d\n";
static char *cantAccessFileMsg	= "%s -- can't access file %s\n";
static char *badLiteralPosMsg   = "%s -- error parsing angles\n";

static readInit ();
static endOfInput();
static lineNumber();
static nextToken();
static readDouble();
static readIdentifier();

static readInit (fp)
FILE *fp;
{
	stream = fp;
	textp = textBuf;
	textBuf[0] = '\0';
	lineNo = 0;
	atEOF = 0;
}

static endOfInput()
{
	return (atEOF);
}

static lineNumber()
{
	return (lineNo);
}

static nextToken()
{
	while (1)
	 { while (isspace(*textp))
	    { textp++;
	    }
	   if (*textp == '\0')
	    { textp = textBuf;
	      if (fgets (textBuf, 256, stream) == NULL)
	       { atEOF = 1;
		 *textp = '\0';
		 return;
	       }
	      else
	       { char *cp;
		 if (cp = (char*)charIndex (textp, '#'))
		  { *cp = '\0';
		  }
		 lineNo++;
	       }
	    }
	   else
	    { return;
	    }
	 }
}

static readDouble(dp)
double *dp;
{
	char *cp;

	if (atEOF)
	 { return (0);
	 }
	nextToken();
	if (cp = (char*)scanDouble (dp, textp))
	 { textp = cp;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static readIdentifier(result, size)
char *result;
int size;
{
	char *cp;

	if (atEOF)
	 { return (0);
	 }
	nextToken();
	if (cp = (char*)scanIdentifier (result, size, textp))
	 { textp = cp;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

int parseLiteralPos (angles, spec, jls)
float *angles;
char *spec;
JLS *jls;
{
	char *fxnName = "parseLiteralPos()";
	char *cp, *ep;
	int j;

	cp = spec;
	for (j=0; j<jls->numJ; j++)
	 { angles[j] = strtod (cp, &ep);
	   if (cp == ep)
	    { errorMsg (ESyntax, "%s -- error reading joint %d\n", 
				  fxnName, j+1);
	      return (-1);
	    }
	   cp = ep;
	   if (jls->jtype[j] == 'R')
	    { angles[j] *= DEGTORAD;
	    }
	   if (j<jls->numJ-1)
	    { ep = cp;
	      while (isspace(*ep))
	       { ep++;
	       }
	      if (*ep == ',')
	       { ep++;
	       }
	      while (isspace(*ep))
	       { ep++;
	       }
	      if (cp == ep)
	       { errorMsg (ESyntax, 
"%s -- ',' or <space> expected after joint %d\n", fxnName, j+1);
		 return (-1);
	       }
	      cp = ep;
	    }
	 }
	return 1;
}

getRobotPosition (jnts, name, robotName)
float *jnts;
const char *name, *robotName;
{
	char *confDir;
	char *fileName;
	char baseName[128];
	char ibuf[256];
	JLS *jls;
	FILE *fp;	       
	char *fxnName = "getRobotPosition()";
	int positionFound = 0;
#if 0	
	enum { READ_NAME,
	       READ_ANGLES } state;
#endif
	if (!planningLevelCheck(fxnName))
	 { return (-1);
	 }
	if ((jls = getJls(robotName)) == NULL)
	 { errorMsg (getErrorCode(0) == EMemory ? EMemory : EBadAccess,
		     rbtCantGetJLSMsg, fxnName, robotName);
	   return (-1);
	 }
	if (strchr ("0123456789-+", name[0]))
	 { if (parseLiteralPos (jnts, name, jls) < 0)
	    { errorMsg (getErrorCode(0), badLiteralPosMsg, fxnName);
	      return (-1);
	    }
	   else
	    { return jls->numJ;
	    }
	 }
	sprintf (baseName, "robotData/%s.pos", robotName);
	if ((fp = locateFile (&fileName, baseName)) == NULL)
	 { errorMsg (EBadAccess, cantAccessFileMsg, fxnName, baseName);
	   return (-1);
	 }

	readInit (fp);

	while (1)
	 { int j;
	   double dval;
	   float deg[MAXJNTS];

	   if (!readIdentifier (ibuf, 256))
	    { if (endOfInput())
	       { break;
	       }
	      else
	       { errorMsg (EBadObj, expectedItemMsg, fxnName, 
				    fileName, "position name", lineNumber());
		 free (fileName);
		 fclose (fp);
		 return (-1);
	       }
	    }
	   for (j=0; j<jls->numJ; j++)
	    { 
	      if (!readDouble (&dval))
	       { errorMsg (EBadObj, expectedItemMsg, fxnName,
				    fileName, "angle values", lineNumber());
		 free (fileName);
		 fclose (fp);
		 return (-1);
	       }
	      deg[j] = dval;
	    }
	   if (strcmp (ibuf, name) == 0)
	    { degreeToRadian (jnts, deg, jls);
	      positionFound = 1;
	    }
	 }
	free (fileName);
	fclose(fp);
	return (positionFound ? jls->numJ : 0);
}

#if TEST

char robotName[256];
char positionName[256];

main()
{
	JLS *jls;
	float jnts[MAXJNTS];
	int status;

	while (1)
	 { printf ("robot and position name> ");
	   scanf ("%s%s", robotName, positionName);
	   status = getRobotPosition (jnts, positionName, robotName);
	   jls = getJls(robotName);
	   if (status < 0)
	    { printErrors();
	      clearErrors();
	    }
	   else if (status == 0)
	    { printf ("no position defined\n");
	    }
	   else
	    { radianToDegree (jnts, jnts, jls);
	      printVf ("%10.4f\n", jnts, status);
	    }
	 }
}

#endif
