#ifndef _TGEN_H_
#define _TGEN_H_

#ifndef PROTO
#ifndef NO_PROTOTYPES
#define PROTO(x) x
#else
#define PROTO(x) ()
#endif
#endif

#include <cdefs.h>
#include <robotMath.h>
#include <robotTypes.h>
#include <robot_kynvar.h>

#include <varargInc.h>

#undef TC		/* ??? defined somewhere in Tsp headers */

#define NJNTS	6	/* XXX should be a parameter */

/* "small" translational and rotational distances */
#define TGEN_MM_EPSILON		(0.001)
#define TGEN_ANG_EPSILON	(2*PI*1e-6)

/* Motion status */

#define TGEN_PENDING	1	/* motion request on queue */
#define TGEN_ACTIVE	2	/* motion being serviced */
#define TGEN_COMPLETED	3	/* motion completed */

typedef struct {
	/* Vector of joint values, wrapped up in a structure for
	   easy handling. */
	float v[NJNTS];		
} JntVec;
	
typedef struct _motionRequest {
	struct _motionRequest *next;
	struct _motionRequest *prev;

	/* Information used to specify motion. Not all items here
	   are used for any given motion. */

#define TGEN_CART_PATH	1	/* Cartesian interpolated motion */
#define TGEN_JNT_PATH	2	/* joint interpolated motion */
	int pathType;		/* path interpolation type */
#define TGEN_CART_TARG	1	/* Cartesian motion target */
#define TGEN_JNT_TARG	2	/* joint angle target */
	int targType;		/* type of target for the motion */

	/* Cartesian motion target info */
	TRSF TC;		/* target position */
	TRSF TL;		/* TOOL transform */
	TRSF TB;		/* start position */
	unsigned long conf;	/* robot configuration */

	/* joint motion target info */
	JntVec jc;		/* joint angles */

	DSPL cdrive;		/* Cartesian drive parameters to the target */
	JntVec jdrive;		/* joint drive parameters to the target */
		
	/* max velocity and accelerations for Cartesian interpolated motions*/
	double maxTransVel;	/* max. Cartesian translational velocity */
	double maxTransAcc;	/* max. Cartesian translational acceleration */
	double maxRotVel;	/* max. Cartesian rotational velocity */
	double maxRotAcc;	/* max. Cartesian rotational acceleration */

	/* max. velocity and accelerations for joint interpolated motions */
	JntVec maxJntVel;	/* max. joint velocity */
	JntVec maxJntAcc;	/* max. joint acceleration */

	/* set the next two fields to explicitly specify the path segment
	   and transition times; set to -1 otherwise. */
	double segTime;		/* path travel time */
	double transTime;	/* transition time for starting this motion */

#define TGEN_IDLE	0x01	/* this is an idle request */
#define TGEN_STOP	0x02	/* this is a stop request */
	int flags;		/* rqeuest information flags */
	int id;			/* request identification number */
} MotionRequest;

typedef struct {
	/* Auxiliary structure used by the API in filling out
	   motion requests. */

	TRSF T6C;
	unsigned long conf;
	JntVec jc;
	unsigned long confReq;
	int movecPathType;
} RequestControl;

typedef struct {
	/* Strutrue used by the trajectory generator to control a
	   motion. When TGEN gets a motion request off the queue, it
	   fills out one of these for the motion.

	   Not all of these fields are filled out at once --
	   some are computed or modified as the motion moves
	   from being next-pending to pending to currently
	   executing. */

	MotionRequest *req;	/* pointer to the request */

	/* The next set of variables control the motion timing. At
	   this point, time has been discretized into the corresponding
	   number of trajectory generator cycles */
	int transCnt;		/* transition time */
	int segCnt;		/* path segment time */
	int segCntEst;		/* estimate of path segment time */

	int cnt;		/* motion time (which is advanced as
				   the motion progresses) */
	double scale;		/* varies (roughly) from 0 to 1 as
				   the motion progresses */
	int status;		/* motion status */

	/* velocity information for joint interpolated motions */
	JntVec jvel;		/* velocity for this motion */
	JntVec prevJvel;	/* velocity for previous motion */
	JntVec delJvel;		/* jvel - prevVel */

	/* velocity information for Cartesian interpolated motions */
	DIFF cvel;		/* velocity for this motion */
	DIFF prevCvel;		/* velocity for previous motion */
	DIFF delCvel;		/* cvel - prevCel */
} MotionControl;

typedef struct _tgenDiag {
	/* structure for logging diagnostic output information */

	struct _tgenDiag *next;
	struct _tgenDiag *prev;
	TRSF T6;
	JntVec j6;
	unsigned long conf;
	MotionControl *mctl;
} TgenDiag;

typedef struct {
	/* Basic trajectory generator control structure. */

	double sampInterval;	/* trajectory generator sample interval */

#define TGEN_ROBOT_ATTACHED	0x01	/* this tgen connected to a robot */
#define TGEN_DIAG_OUTPUT	0x02	/* produce diagnostic output */
	int flags;		/* control flags */

	/* motion request queue. Add at the tail; remove at the head. */
	MotionRequest *reqHead;
	MotionRequest *reqTail;

	MotionRequest *queueTail;

	/* trajectory generator keep track of four motions at once:
	   the previous, current, pending, and next pending. */
	MotionControl *prevMot;
	MotionControl *currMot;
	MotionControl *nextMot;
	MotionControl *nextnextMot;

	/* run-time information: */

	int timeCnt;		/* cycle counter */

	/* output setpoint values produced by the trajectory generator */
	TRSF T6;		/* T6 setpoint */
	unsigned long conf;	/* robot configuration */
	JntVec j6;		/* j6 setpoint */
	JntVec oldJ6;		/* previous j6 setpoint */

	/* actual positions read back from the robot */
	TRSF T6real;		/* real T6 */
	unsigned long confReal;	/* real configuration */
	JntVec j6real;		/* real j6 (joint values) */
	
	FORCE force;		/* 6 DOF wrist force sensor info */
	TRSF Alter;		/* used for on-line trajectory modification*/

	/* robot information */

	char *name;		/* name of the robot */
	JLS *jls;		/* joint-level information structure */
	KYN *kyn;		/* kinematic/dynamic information structure */
	void *var;		/* robot-specific scratch space (not used) */
	
	/* Api information */

	MotionRequest reqTmpl;	/* template used by the API to create
				   motion requests */
	RequestControl reqCtrl;

	/* diagnostic information */

	TgenDiag *diagHead;
	TgenDiag *diagTail;

	/* real-time control info. These are pointers to routines
	   which are implementation-specific. */

	int lockspace;		/* scratch space for lock routines */
	void (*lock)();		/* lock access to trajectory generator */
	void (*unlock)();	/* unlock access */
	void (*abort)();	/* abort the program */
	void *ctrl;		/* implementation-specific storage */
	int (*alterFxn)();	/* on-line trajectory modification fxn */
	int alterArg;		/* integer and */
	void *alterPtr;		/*   pointer arguments for alterFxn */
	int (*monitorFxn)();	/* real-time monitor function */
	int monitorArg;		/* integer and */
	void *monitorPtr;	/*   pointer arguments for monitor fxn */
} Tgen;

/* error codes */

#define TGBadArg		1
#define TGNoRobotJls		2
#define TGNoRobotKyn		3
#define TGNoJntMotion		4
#define TGNoCartMotion		5
#define TGJntLimit		6
#define TGWorkspaceLimit	7
#define TGVelLimit		8
#define TGSingular		9
#define TGComError	       10

Tgen *tgenAlloc PROTO((char*));
void tgenFree PROTO((Tgen*));
int tgenInitJnt PROTO((Tgen*, double, JntVec*));
int tgenInitCart PROTO((Tgen*, double, TRSF*, unsigned long));
int tgenSetPath PROTO((Tgen*, int));
int tgenSetConf PROTO((Tgen*, unsigned long));
int tgenSetCartVel PROTO((Tgen*, double, double));
void tgenGetCartVel PROTO((Tgen*, double*, double*));
int tgenSetJntVel PROTO((Tgen*, double, int));
double tgenGetJntVel PROTO((Tgen*, int));
int tgenSetCartAcc PROTO((Tgen*, double, double));
void tgenGetCartAcc PROTO((Tgen*, double*, double*));
int tgenSetJntAcc PROTO((Tgen*, double, int));
double tgenGetJntAcc PROTO((Tgen*, int));
int tgenSetMotionTime PROTO((Tgen*, double, double));
int tgenMoveC PROTO((Tgen*, TRSF*, TRSF*));
int tgenMoveJ PROTO((Tgen*, JntVec*));
int tgenStop PROTO((Tgen*, double));
int tgenIsBusy PROTO((Tgen*));
int tgenRun PROTO((Tgen*, int));
void tgenEnableDiag PROTO((Tgen*, int));
void tgenClearDiag PROTO((Tgen*));
TgenDiag *tgenGetDiag PROTO((Tgen*));
int (*tgenAlterFxnSet PROTO((Tgen*, int (*)(), int, void*)))();
int (*tgenMonitorFxnSet PROTO((Tgen*, int (*)(), int, void*)))();
void tgenAbort VA_PROTO((Tgen *tg, int code, char *fmt, ...));
/* Internal support routines */

void computeLinearMotionCart PROTO((TRSF*, TRSF*, DSPL*, double));
void computeLinearMotionJnt PROTO((JntVec*, JntVec*, JntVec*, double));

#endif
