#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/edt.h>
#include <sys/vmereg.h>
#include <sys/pio.h>

#include "rciioctl.h"
#include "xydreg.h"

#define STATIC

static unsigned char idcheck[] = "VMEIDXYC240    1 11 ";

int xyddevflag = 0;
extern int xyd_internal_major;
extern XydSoftc xydi[];
static int goodid();

void xydintr (int unit)
{
	/* Called in response to an interrupt from the XYD */

	XydSoftc *info;
	int pending;
	int s;

	info = &xydi[unit];

	while (pending = info->xyd->intpend)
	 { info->xyd->intclear = pending;

	   if (info->flags & XYD_BUSY)	/* is device open? */
	    { if (pending & XYD_INTRMASK)
	       { 
		 s = splhi();
		 if (info->intrCnt < info->qsize)
		  { info->intrCnt++;		/* bump the counter */
		    pollwakeup (info->pollhead, POLLRDNORM);
		  }
		 splx(s);
		 pending &= ~(XYD_INTRMASK);
	       }
	    }
	   if (pending)
	    { /* declare a stray interrupt */
	      cmn_err (CE_CONT, "xyd%d: stray interrupt(s); mask=0x%x\n",
				 unit, pending);
	    }
	 }
}

static int goodid (Xyd *xyd)
{
	/* Verify that the board is really an XVME-240, using the
	   ID signature built in to the board. */

	int i;

	for (i=0;i<20;i++)
	 { if (badaddr(&xyd->id[i+i+1],1))
	    { cmn_err (CE_CONT, "xyd0: not present");
	      return(0);
	    }
	   if (xyd->id[i+i+1] != idcheck[i])
	    { cmn_err (CE_CONT, "xyd0: not an xyd");
	      return(0);
	    }
	 }
	return(1);
}

STATIC void xydedtinit (edt_t *e)
{
	/* Initialization routine. Verify that the board is there, and
	   initialize its registers to something reasonable. */

	int i, vec;
	Xyd *xyd;
	piomap_t *pmap;

	pmap = pio_mapalloc(e->e_bus_type, e->e_adap, &e->e_space[0],
			    PIOMAP_FIXED, "xyd");
	if (pmap == 0)
	 { return;
	 }
	xyd = (Xyd*)pio_mapaddr (pmap, e->e_iobase);
	if (! goodid(xyd))
	 { pio_mapfree (pmap);
	   return;
	 }
	xydi[e->e_ctlr].xyd = 0;

	vec = vme_ivec_alloc (e->e_adap);
	if (vec == -1)
	 { cmn_err (CE_WARN, "xyd: unit %d, no irq vector\n", e->e_ctlr);
	   pio_mapfree (pmap);
	   return;
	 }

	vme_ivec_set (e->e_adap, vec, xydintr, e->e_ctlr);

	xyd->csr = XYD_RESET;
	xyd->csr = 0;
	xyd->csr = XYD_RED_OFF | XYD_GREEN_OFF;
	xyd->dirs = XYD_DIRMASK;
	xyd->vector = vec;
	xydi[e->e_ctlr].xyd = xyd;
	cmn_err (CE_CONT, "xydedtinit: unit %d vector %x ipl %d\n",
		e->e_ctlr, vec, ((vme_intrs_t*)(e->e_bus_info))->v_brl);
}

STATIC int xydopen (dev_t *dev, int flag, int otyp, cred_t *crp)
{
	/* This device is an exclusive open device: only one process
	   may open it at a time. This prevents contention problems
	   and makes the driver software easier to implement. */

	XydSoftc *info;
	Xyd *xyd;
	int oldflags;
	int s;

	info = &xydi[geteminor(*dev)];

	oldflags = info->flags;

	if (! info->xyd)
	 { return (ENXIO);
	 }
	xyd = info->xyd;
	s = splhi();
	if (info->flags & XYD_BUSY)
	 { splx(s);
	   return (EBUSY);
	 }
	else
	 { info->flags |= XYD_BUSY;
	 }
	splx(s);

	info->qsize = TQ_DEFAULT_SIZE;
	info->intrCnt = 0;
	info->pollhead = phalloc (KM_SLEEP);

	xyd->dirs = XYD_DIRMASK;
	xyd->intclear = XYD_INTRCLEAR;
	return (0);
}

static void doclose (XydSoftc *info)
{
	Xyd *xyd;
	int s;

	xyd = info->xyd;

	xyd->csr = XYD_RED_OFF | XYD_GREEN_OFF;
	xyd->intmask = 0;
	xyd->intclear = 0xff;
	if (info->pollhead)
	 { phfree (info->pollhead);
	 }
	else
	 { cmn_err (CE_WARN, "xydclose: pollhead is NULL\n");
	 }
	info->flags = 0;
}

STATIC int xydclose (dev_t dev, int flag, int otyp, cred_t *crp)
{
	/* When closing the device, reset the registers to what
	   they were when the device was openned. */

	XydSoftc *info;

	info = &xydi[geteminor(dev)];

	/* check for soft-close */
	if (! (info->flags & XYD_BUSY) )
	 { /* don't return an error, since this may happen normally */
/*DBG*/cmn_err (CE_CONT, "hard close returned, nothing to do\n");
	   return (0);
	 }
	doclose (info);
/*DBG*/cmn_err (CE_CONT, "hard close called\n");
	return (0);
}

STATIC int xydread (dev_t dev, uio_t *uiop, cred_t *crp)
{
	/* The read routine returns a single byte (set to 1) if an interrupt
	   arrived. */

	XydSoftc *info;
	int cnt, s;
	int ack = 1;

	info = &xydi[geteminor(dev)];

	/* check for soft-close */
	if (! (info->flags & XYD_BUSY) )
	 { return (EBADF);
	 }

	s = splhi();
	cnt = info->intrCnt;
	if (cnt)
	 { info->intrCnt--;
	 }
	splx(s);
	if (cnt)
	 { if (ureadc (ack, uiop) != 0)
	    { return (EFAULT);
	    }
	 }
	return (0);
}

STATIC int xydwrite (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STATIC int xydchpoll (dev_t dev, short events, int anyyet,
		      short *reventsp, struct pollhead **phpp)
{
	/* Check to see if any interrupt events have arrived. If not,
	   then queue the process for a possible select wakeup. */

	XydSoftc *info;
	int cnt, s;

	info = &xydi[geteminor(dev)];

	/* check for soft-close */
	if (! (info->flags & XYD_BUSY) )
	 { return (EBADF);
	 }

	/* What about races? */
	if ((events & (POLLIN|POLLRDNORM)) && info->intrCnt > 0)
	 { *reventsp = events & (POLLIN|POLLRDNORM);
	 }
	else
	 { *reventsp = 0;
	   if (!anyyet)
	    { *phpp = info->pollhead;
	    }
	 }
	return (0);
}

STATIC int xydioctl (dev_t dev, int cmd, void *arg, int mode, 
		     cred_t *crp, int *rvalp)
{
	XydSoftc *info;

	info = &xydi[geteminor(dev)];

	/* check for soft-close */
	if (! (info->flags & XYD_BUSY) )
	 { return (EBADF);
	 }

	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case RCIIOC_CONFIRM:
	    { break;
	    }
	   case RCIIOC_INTR_ENABLE:
	    { int mask = (int)arg;
	      Xyd *xyd = info->xyd;
	      int s;

	      s = splhi(); /* is this really necessary? */
	      if (mask)
	       { xyd->intmask |= XYD_INTRMASK;
		 xyd->intclear = XYD_INTRCLEAR;
		 if (! (xyd->csr & XYD_INTENB))
		  { xyd->csr |= XYD_INTENB;
		  }
	       }
	      else
	       { xyd->csr &= ~XYD_INTENB;
		 xyd->intmask = 0;
		 xyd->intclear = 0xff;
	       }
	      splx(s);
	      break;
	    }
	   case RCIIOC_SET_INTRQ:
	    { int size, s;

	      size = (int)arg;
	      if (size <= 0)
	       { return (EINVAL);
	       }
	      else
	       { s = splhi();
		 info->qsize = size;
		 info->intrCnt = 0;
		 splx(s);
	       }
	      break;
	    }
	   case RCIIOC_GET_INTRQ:
	    { int size;

	      size = info->qsize;
	      if (copyout ((caddr_t)&size, (caddr_t)arg, sizeof(int)))
	       { return (EFAULT);
		 break;
	       }
	      break;
	    }
/* RCIIOC_SOFT_CLOSE is a special hack just for the SGI machines, since
   it seems there is a bug: calling close(2) does not always result
   in the driver close routine being called right away. So we fake
   it with an ioctl, just to be sure. What a botch! */
	   case RCIIOC_SOFT_CLOSE:
	    { doclose (info);      
/*DBG*/cmn_err (CE_CONT, "soft close called\n");
	      break;
	    }
	 }
	return (0);
}
