#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/sysmacros.h>
#include <sys/stat.h>

#include "s16dreg.h"
#include "s16dioctl.h"
#include "rcisys.h"
#include "rciioctl.h"

#define STATIC

int s16ddevflag = 0;
void *s16di;

#define S16D_INTR_SPEC	0	/* device interrupt spec. */
#define S16D_REG_SPEC	0	/* register set to be mapped */
#define S16D_NULL_HANDLER (u_int(*)(caddr_t))nulldev

/* maximum number of instances supported by this driver */
#define MAX_INSTANCES		4
/* number of bits in the minor node reserved for the instance number */
#define NUM_INSTANCE_BITS	4
/* returns the instance number from a dev_t structure */
#define GET_INSTANCE(d)		(getminor(d) & ((1<<NUM_INSTANCE_BITS)-1))
/* bit setting or-ed with the instance number to get an mmap device */
#define MMAP_MINOR		(1<<NUM_INSTANCE_BITS)
/* returns true if the minor number corresponds to an mmap device */
#define IS_MMAP_DEV(m)	(((m) & ~((1<<NUM_INSTANCE_BITS)-1)) == MMAP_MINOR)

/* Hardware-specific device routines */
static void s16dHardwareReset (S16dRegs*, S16dSoftc*);
static void s16dHardwareIntrSetup (S16dRegs*, S16dSoftc*);
static void s16dHardwareOpen (S16dRegs*, S16dSoftc*);
static void s16dHardwareClose (S16dRegs*, S16dSoftc* );
static void s16dHardwareIntrEnable (S16dRegs*, S16dSoftc*, int);
static  int s16dHardwareIntrHandle (S16dRegs*, S16dSoftc*, int*);
static u_short s16dGetCmdReg (S16dRegs*, S16dSoftc*);

/* Stuff relating to dynamic loading and autoconfiguration. There
   is *a lot* of boiler plate associated with this. */

static int s16didentify();
static int s16dattach();
static int s16ddetach();
static int s16dgetinfo();

static int s16dopen(dev_t *devp, int flag, int otyp, cred_t *credp);
static int s16dclose(dev_t dev, int flag, int otyp, cred_t *credp);
static int s16dread(dev_t dev, struct uio *uiop, cred_t *credp);
static int s16dwrite(dev_t dev, struct uio *uiop, cred_t *credp);
static int s16dioctl(dev_t dev, int cmd, int arg, int mode, 
			cred_t *credp, int *rvalp);
static int s16dmmap(dev_t dev, off_t off, int prot);
static int s16dchpoll(dev_t dev, short events, int anyyet,
		    short *reventsp, struct pollhead **phpp);

static u_int s16dintr (caddr_t);
static u_int s16dsoftintr (caddr_t);

static struct cb_ops s16d_cb_ops = {
	s16dopen,
	s16dclose,
	nodev,
	nodev,
	nodev,
	s16dread,
	s16dwrite,
	s16dioctl,
	nodev,
	s16dmmap,
	nodev,
	s16dchpoll,
	nodev,
	NULL,
	D_NEW | D_MP
};

static struct dev_ops s16d_ops = {
	DEVO_REV,
	0,
	s16dgetinfo,
	s16didentify,
	nulldev,
	s16dattach,
	s16ddetach,
	nodev,
	&s16d_cb_ops,
	(struct bus_ops *)NULL
};

static struct modldrv modldrv = {
	&mod_driverops,
	"s16d driver",
	&s16d_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	NULL
};

int _init (void)
{
	int status;

	ddi_soft_state_init (&s16di, sizeof(S16dSoftc), MAX_INSTANCES);
	status = mod_install (&modlinkage);
	if (status != 0)
	 { ddi_soft_state_fini (&s16di);
	 }
	return (status);
}

int _fini (void)
{
	int status;

	status = mod_remove (&modlinkage);
	if (status == 0)
	 { ddi_soft_state_fini (&s16di);
	 }
	return (status);
}

int _info (struct modinfo *modinfop)
{
	return (mod_info (&modlinkage, modinfop));
}

static int s16didentify (dip)
dev_info_t *dip;
{
	/* For SBus devices, the name checked by this routine must match
	   the name stored on the device itself. */

	if (strcmp (ddi_get_name(dip), "EDT,s16d") == 0)
	 { return (DDI_IDENTIFIED);
	 }
	else
	 { return (DDI_NOT_IDENTIFIED);
	 }
}

static int s16dgetinfo (dev_info_t *dip, ddi_info_cmd_t infocmd, 
			  void *arg, void **result)
{
	S16dSoftc *info;
	int status;

	switch (infocmd)
	 { case DDI_INFO_DEVT2INSTANCE:
	    { *result = (void*)GET_INSTANCE((dev_t)arg);
	      status = DDI_SUCCESS;
	      break;
	    }
	   case DDI_INFO_DEVT2DEVINFO:
	    { info = ddi_get_soft_state(s16di, GET_INSTANCE((dev_t)arg));
	      if (info == NULL)
	       { return (DDI_FAILURE);
	       }
	      *result = (void*)info->dip;
	      status = DDI_SUCCESS;
	      break;
	    }
	   default:
	    { status = DDI_FAILURE;
	      break;
	    }
	 }
	return (status);
}

static int s16dfreeResources (dev_info_t *dip, S16dSoftc *info)
{
	if ((info->resources & S16D_MMAP_MINOR_CREATED) ||
	    (info->resources & S16D_MAIN_MINOR_CREATED))
	 { ddi_remove_minor_node (dip, NULL);
	 }
	if (info->resources & S16D_IHANDLER_SET)
	 { ddi_remove_intr (dip, S16D_INTR_SPEC, &info->iblkCookie);
	 }
	if (info->resources & S16D_SOFT_IHANDLER_SET)
	 { ddi_remove_softintr (info->softIntrId);
	 }
	if (info->resources & S16D_LOW_MUTEX_SET)
	 { mutex_destroy (&info->lowMux);
	 }
	if (info->resources & S16D_INTR_MUTEX_SET)
	 { mutex_destroy (&info->intrMux);
	 }
	if (info->resources & S16D_REGS_MAPPED)
	 { ddi_unmap_regs (dip, S16D_REG_SPEC, (caddr_t*)&info->regs,
		      		   /*offset=*/0, /*len=*/0);
	 }
	ddi_soft_state_free (s16di, info->instance);
}

static int s16dattach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	S16dSoftc *info;
	int instance;

	static char *mainNames[] = { "0", "1", "2", "3" };
	static char *mmapNames[] = { "mem0", "mem1", "mem2", "mem3" };
printf ("s16dattach()\n");
	switch (cmd)
	 { case DDI_ATTACH:
	    { instance = ddi_get_instance(dip);
	      if (instance >= MAX_INSTANCES)
	       { return (DDI_FAILURE);
	       }
	      if (ddi_soft_state_zalloc(s16di, instance) != 0)
	       { return (DDI_FAILURE);
	       }
	      info = ddi_get_soft_state(s16di, instance);

	      info->dip = dip;
	      info->instance = instance;
	
	      /* map device registers */
	      if (ddi_map_regs (dip, S16D_REG_SPEC, (caddr_t*)&info->regs,
				/*offset=*/0, /*len=*/0) != DDI_SUCCESS)
	       { s16dfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= S16D_REGS_MAPPED;

	      /* make device quiescent */
	      s16dHardwareReset (info->regs, info);

	      /* set up interrupt-level mutex for the device. Start
		 by assigning a dummy interrupt handler so the
		 mutex is associated with the proper interrupt context,
		 without risk of an a handler getting called. */

	      if (ddi_add_intr (dip, S16D_INTR_SPEC,&info->iblkCookie,
		     NULL, S16D_NULL_HANDLER, NULL) != DDI_SUCCESS)
	       { s16dfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }

	      mutex_init (&info->intrMux, "s16d intr mutex", MUTEX_DRIVER,
			  (void*)info->iblkCookie);
	      ddi_remove_intr (dip, S16D_INTR_SPEC, info->iblkCookie);
	      info->resources |= S16D_INTR_MUTEX_SET;

	      /* see if the device interrupt is delivered in a high-level
		 interrupt context. If it is, we need to set up a 
		 low-level interrupt handler and mutex as well,
		 because we can't call cmn_err() or pollwakeup()
		 from a high-level interrupt context. */

	      if (ddi_intr_hilevel (dip, S16D_INTR_SPEC))
	       { info->hasHighLevelIntr = 1;
		 
		 /* set up low-level mutex and handler */

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			S16D_NULL_HANDLER, NULL) != DDI_SUCCESS)
		  { s16dfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 mutex_init (&info->lowMux, "s16d low mutex", MUTEX_DRIVER,
			     (void*)info->softIblkCookie);
		 ddi_remove_softintr (info->softIntrId);
		 info->resources |= S16D_LOW_MUTEX_SET;

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			s16dsoftintr, (caddr_t)info) != DDI_SUCCESS)
		  { s16dfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 info->resources |= S16D_SOFT_IHANDLER_SET;

		 info->muxp = &info->lowMux;
		 cmn_err (CE_CONT, "s16d: interrupts are high level\n");
	       }
	      else
	       { info->hasHighLevelIntr = 0;
		 info->muxp = &info->intrMux;
		 cmn_err (CE_CONT, "s16d: interrupts are low level\n");
	       }

	      if (ddi_add_intr (dip, S16D_INTR_SPEC, &info->iblkCookie,
		     &info->idevCookie, s16dintr,(caddr_t)info) != DDI_SUCCESS)
	       { s16dfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= S16D_IHANDLER_SET;

	      /* program device interrupt vector */
	      s16dHardwareIntrSetup (info->regs, info);

	      if (ddi_create_minor_node (dip, mainNames[instance], S_IFCHR, 
		     instance, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { s16dfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= S16D_MAIN_MINOR_CREATED;

	      if (ddi_create_minor_node (dip, mmapNames[instance], S_IFCHR, 
		     instance|MMAP_MINOR, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { s16dfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= S16D_MMAP_MINOR_CREATED;

	      ddi_report_dev(dip);
cmn_err (CE_CONT, "succeeded\n");
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }
}

static int s16ddetach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	S16dSoftc *info;
	
	switch (cmd)
	 { case DDI_DETACH:
	    { info = ddi_get_soft_state (s16di, ddi_get_instance(dip));

	      /* make device quiescent */
	      s16dHardwareReset (info->regs, info);
	      s16dfreeResources (dip, info);
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }	
}

static u_int s16dintr (caddr_t arg)
{
	/* Called in response to an interrupt from the S16D */

	S16dSoftc *info = (S16dSoftc*)arg;
	int needSoftIntr = 0;
	int pending;
	int stray;

	mutex_enter (&info->intrMux);
	pending = s16dHardwareIntrHandle (info->regs, info, &stray);
	if (pending == 0 && stray == 0)
	 { mutex_exit (&info->intrMux);
	   return (DDI_INTR_UNCLAIMED);
	 }

	if (pending)
	 { if (info->hasHighLevelIntr)
	    { if (info->highIntrCnt < info->qsize)
	       { info->highIntrCnt++;
		 if (!info->softIntrRunning)
		  { needSoftIntr = 1;
		  }
	       }
	    }
	   else
	    { if (info->intrCnt < info->qsize)
	       { info->intrCnt++;		/* bump the counter */
		 mutex_exit (&info->intrMux);
		 pollwakeup (&info->pollhead, POLLRDNORM);
		 mutex_enter (&info->intrMux);
	       }
	    }
	 }
	if (stray)
	 { /* declare a stray interrupt */
	   if (info->hasHighLevelIntr)
	    { info->stray = stray;
	      if (!info->softIntrRunning)
	       { needSoftIntr = 1;
	       }
	    }
	   else
	    { cmn_err (CE_CONT, "s16d%d: stray interrupt(s); mask=0x%x\n",
		   		 info->instance, stray);
	    }
	 }
	mutex_exit (&info->intrMux);
	if (needSoftIntr)
	 { ddi_trigger_softintr (info->softIntrId);
	 }
	return (DDI_INTR_CLAIMED);
}

static u_int s16dsoftintr (caddr_t arg)
{
	/* Used only when the device interrupts occur at high interrupt
	   priority. Calls pollwakeup and cmn-err because the high-level
	   interrupt handler can't. */

	S16dSoftc *info = (S16dSoftc*)arg;
	int stray;
	int cnt;

	mutex_enter (&info->lowMux);
	mutex_enter (&info->intrMux);

	/* don't know why info->softIntrRunning needs to be checked
	   in the next phrase, but it was in the manual example */
	if ((info->stray == 0 && info->highIntrCnt <= 0) ||
	    info->softIntrRunning)
	 { mutex_exit (&info->intrMux);
	   mutex_exit (&info->lowMux);
	   return (DDI_INTR_UNCLAIMED);	
	 }
	info->softIntrRunning = 1;
	while (info->stray || info->highIntrCnt > 0)
	 { /* ASSERT (mutex_owned (&info->intrMux)); */

	   if ((stray = info->stray) != 0)
	    { info->stray = 0;
	    }
	   if ((cnt = info->highIntrCnt) > 0)
	    { info->highIntrCnt--;
	    }
	   mutex_exit (&info->intrMux);

	   if (stray)
	    { cmn_err (CE_CONT, "s16d%d: stray interrupt(s); mask=0x%x\n",
				 info->instance, stray);
	    }
	   if (cnt)
	    { if (info->intrCnt < info->qsize)
	       { info->intrCnt++;
		 mutex_exit (&info->lowMux);
		 pollwakeup (&info->pollhead, POLLRDNORM);
		 mutex_enter (&info->lowMux);
	       }
	    }

	   mutex_enter (&info->intrMux);
	 }
	info->softIntrRunning = 0;
	mutex_exit (&info->intrMux);
	mutex_exit (&info->lowMux);
	return (DDI_INTR_CLAIMED);	
}

STATIC int s16dopen (dev_t *dev, int flag, int otyp, cred_t *crp)
{
	/* This device is an exclusive open device: only one process
	   may open it at a time. This prevents contention problems
	   and makes the driver software easier to implement. */

	S16dSoftc *info;

	if ((info = ddi_get_soft_state(s16di,GET_INSTANCE(*dev))) == NULL)
	 { return (ENXIO);
	 }
	if (otyp != OTYP_CHR)
	 { return (EINVAL);
	 }
	if (!info->regs)	/* paranoid; this check shouldn't be needed */
	 { return (ENXIO);
	 }

	/* now determine whether this is a main device or an mmap device.
	   Enforce exclusive open for main devices */
	mutex_enter (info->muxp);
	if (IS_MMAP_DEV(getminor(*dev)))
	 { /* OK, just return */
	 }
	else
	 { if (info->flags & S16D_BUSY)
	    { mutex_exit (info->muxp);
	      return (EBUSY);
	    }
	   else
	    { info->flags |= S16D_BUSY;
	    }
	   info->qsize = TQ_DEFAULT_SIZE;
	   info->intrCnt = 0;
	   bzero ((char*)&info->pollhead, sizeof(struct pollhead));

	   s16dHardwareOpen (info->regs, info);
	 }
	mutex_exit (info->muxp);
	return (0);
}

static void doclose (S16dSoftc *info)
{
	s16dHardwareClose (info->regs, info);
	info->flags = 0;
}

STATIC int s16dclose (dev_t dev, int flag, int otyp, cred_t *crp)
{
	S16dSoftc *info;

	if ((info = ddi_get_soft_state(s16di,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if (IS_MMAP_DEV(getminor(dev)))
	 { return (0);
	 }
	/* check for soft-close */
	mutex_enter (info->muxp);
	if (! (info->flags & S16D_BUSY) )
	 { mutex_exit (info->muxp);
	   return (0);
	 }
	mutex_exit (info->muxp);
	doclose (info);
	return (0);
}

STATIC int s16dread (dev_t dev, uio_t *uiop, cred_t *crp)
{
	/* The read routine returns a single byte (set to 1) if an interrupt
	   arrived. */

	S16dSoftc *info;
	int cnt;
	int ack = 1;

	if ((info = ddi_get_soft_state(s16di,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if (IS_MMAP_DEV(getminor(dev)))
	 { return (ENXIO);
	 }
	/* check for soft-close */
	mutex_enter (info->muxp);
	if (! (info->flags & S16D_BUSY) )
	 { mutex_exit (info->muxp);
	   return (EBADF);
	 }
	cnt = info->intrCnt;
	if (cnt)
	 { info->intrCnt--;
	 }
	mutex_exit (info->muxp);

	if (cnt)
	 { if (ureadc (ack, uiop) != 0)
	    { return (EFAULT);
	    }
	 }
	return (0);
}

STATIC int s16dwrite (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STATIC int s16dmmap (dev_t dev, off_t off, int prot)
{
	S16dSoftc *info;

	if ((info = ddi_get_soft_state(s16di,GET_INSTANCE(dev))) == NULL)
	 { return (-1);
	 }
	if (off < 0 || off >= sizeof(S16dRegs))
	 { return (-1);
	 }
	return (hat_getkpfnum((char*)info->regs + off));
}

STATIC int s16dchpoll (dev_t dev, short events, int anyyet,
		      short *reventsp, struct pollhead **phpp)
{
	/* Check to see if any interrupt events have arrived. If not,
	   then queue the process for a possible select wakeup. */

	S16dSoftc *info;
	int cnt;

	if ((info = ddi_get_soft_state(s16di,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if (IS_MMAP_DEV(getminor(dev)))
	 { return (ENXIO);
	 }
	/* check for soft-close */
	mutex_enter (info->muxp);
	if (! (info->flags & S16D_BUSY) )
	 { mutex_exit (info->muxp);
	   return (EBADF);
	 }
	cnt = info->intrCnt;
	mutex_exit (info->muxp);

	/* not really sure how system prevents races with the interrupt
	   handler here, but this is what the example says to do ... */
	if ((events & (POLLIN|POLLRDNORM)) && cnt > 0)
	 { *reventsp = events & (POLLIN|POLLRDNORM);
	 }
	else
	 { *reventsp = 0;
	   if (!anyyet)
	    { *phpp = &info->pollhead;
	    }
	 }
	return (0);
}

STATIC int s16dioctl (dev_t dev, int cmd, int arg, int mode, 
		     cred_t *crp, int *rvalp)
{
	S16dSoftc *info;
	int isMmapDev;

	if ((info = ddi_get_soft_state(s16di,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if ((isMmapDev = IS_MMAP_DEV(getminor(dev))) == 0)
	 { /* check for soft-close */
	   mutex_enter (info->muxp);
	   if (! (info->flags & S16D_BUSY) )
	    { mutex_exit (info->muxp);
	      return (EBADF);
	    }
	   mutex_exit (info->muxp);
	 }

	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case RCIIOC_CONFIRM:
	    { break;
	    }
	   case RCIIOC_INTR_ENABLE:
	    { mutex_enter (&info->intrMux);
	      s16dHardwareIntrEnable (info->regs, info, (int)arg);
	      mutex_exit (&info->intrMux);
	      break;
	    }
	   case RCIIOC_SET_INTRQ:
	    { int size;

	      if (isMmapDev)
	       { return (ENOTTY);
	       }
	      size = (int)arg;
	      if (size <= 0)
	       { return (EINVAL);
	       }
	      else
	       { mutex_enter (info->muxp);
		 info->qsize = size;
		 info->intrCnt = 0;
		 mutex_exit (info->muxp);
	       }
	      break;
	    }
	   case RCIIOC_GET_INTRQ:
	    { int size;

	      if (isMmapDev)
	       { return (ENOTTY);
	       }
	      size = info->qsize;
	      if (copyout ((caddr_t)&size, (caddr_t)arg, sizeof(int)))
	       { return (EFAULT);
		 break;
	       }
	      break;
	    }
/* RCIIOC_SOFT_CLOSE is a special hack. On some machines, calling
   close(2) does not always result in the driver close routine being 
   called right away. So we fake it with an ioctl, just to be sure.
   It is the responsiblity of the application to call close(2) directly
   after the SOFT_CLOSE. Once SOFT_CLOSE has been called, the driver 
   will allow another exclusive open on the main device, and so concurrent
   accesses by the original descriptor are possible (but probably not
   desirable). */   

	   case RCIIOC_SOFT_CLOSE:
	    { if (isMmapDev)
	       { return (ENOTTY);
	       }
	      doclose (info);      
	      break;
	    }
/* This next bunch of ioctls are specific to the S16D, and provide
   access to the control and status regsiters. Note that the command
   register is write-only. Some of it's bits are momentary, and others
   can be read back (either directly or in inverted form) from the status
   register. However, DISINT is stored by the driver, and so if it is
   set outside the driver interface, the driver will be unaware of this.

   Similarly, the config register is write-only, and so all of it's
   bits are stored by the driver.
 */
	   case S16D_GET_COMMAND:
	    { u_short cmd;

	      mutex_enter (info->muxp);
	      cmd = s16dGetCmdReg (info->regs, info);
	      mutex_exit (info->muxp);
	      if (copyout ((caddr_t)&cmd, (caddr_t)arg, sizeof(u_short)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case S16D_SET_COMMAND:
	    { u_short cmd;

	      if (copyin ((caddr_t)arg, (caddr_t)&cmd, sizeof(u_short)))
	       { return (EFAULT);
	       }
	      mutex_enter (info->muxp);
	      info->regs->csr = cmd;
	      info->intrEnable = !(cmd & S16D_DISINT);
	      mutex_exit (info->muxp);
	    }
	   case S16D_GET_STATUS:
	    { u_short status;

	      mutex_enter (info->muxp);
	      status = info->regs->csr;
	      mutex_exit (info->muxp);
	      if (copyout ((caddr_t)&status, (caddr_t)arg, sizeof(u_short)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case S16D_GET_CONFIG:
	    { u_short config;

	      mutex_enter (info->muxp);
	      config = info->config;
	      mutex_exit (info->muxp);
	      if (copyout ((caddr_t)&config, (caddr_t)arg, sizeof(u_short)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case S16D_SET_CONFIG:
	    { u_short config;

	      if (copyin ((caddr_t)arg, (caddr_t)&config, sizeof(u_short)))
	       { return (EFAULT);
	       }
	      mutex_enter (info->muxp);
	      info->config = config;
	      info->regs->config = config;
	      mutex_exit (info->muxp);
	      break;
	    }
	 }
	return (0);
}

static u_short s16dGetCmdReg (S16dRegs *regs, S16dSoftc *info)
{
	u_short status;

	status = regs->csr;
	return ((status & S16D_CMD_STORED) |
		(~(status & S16D_CMD_STORED_I) & S16D_CMD_STORED_I) |
		(info->intrEnable ? 0 : S16D_DISINT) );
}

static void s16dHardwareReset (S16dRegs *regs, S16dSoftc *info)
{
	/* make the device quiescent. Called at attach and detach time */

	regs->csr = S16D_RESET;
	info->intrEnable = 0;
	info->config = 0;
	regs->config = 0;
}

static void s16dHardwareIntrSetup (S16dRegs* regs, S16dSoftc* info)
{
	/* if the device interrupt vector or IPL is programmable,
	   do this here based on info provided in info->idevCookie. */

	/* Nothing to do for S16D */
}

static void s16dHardwareClose(S16dRegs *regs, S16dSoftc *info)
{
	/* make the device quiet when it is closed */

	regs->csr = (s16dGetCmdReg (regs,info) & S16D_FREE_CMD_BITS) |
		     S16D_RESET;
	info->intrEnable = 0;
	info->config &= ~(S16D_ENEOD | S16D_ENDINT);
	regs->config = info->config;
}

static void s16dHardwareOpen(S16dRegs *regs, S16dSoftc *info)
{
	/* initialize the device when it is openned */	

	regs->csr = (s16dGetCmdReg (regs,info) & S16D_FREE_CMD_BITS) |
		     S16D_RESET;
	info->intrEnable = 0;
	info->config &= ~(S16D_ENEOD | S16D_ENDINT);
	regs->config = info->config;
}

static void s16dHardwareIntrEnable(S16dRegs *regs, S16dSoftc *info, int enable)
{
	/* enable/disable device interrupts */

	u_short cmd;

	cmd = s16dGetCmdReg (regs, info);
	if (enable)
	 { regs->csr = (cmd | S16D_REOD | S16D_RDINT);
	   info->config |= S16D_ENDINT;
	   regs->config = info->config;
	   regs->csr = (cmd & ~S16D_DISINT);
	   info->intrEnable = 1;
	 }
	else
	 { regs->csr = (cmd | S16D_DISINT | S16D_REOD | S16D_RDINT);
	   info->config &= ~S16D_ENDINT;
	   regs->config = info->config;
	   info->intrEnable = 0;
	 }
}

static int s16dHardwareIntrHandle(S16dRegs *regs, S16dSoftc *info, int *stray)
{
	/* clear interrupts, return true if a valid interrupt is pending,
	   and return a mask of any stray interrupts in *stray. */
	int pending;

	pending = regs->csr & (S16D_EODMA_S | S16D_INTDEV_S);
	regs->csr = s16dGetCmdReg (regs, info) | S16D_REOD | S16D_RDINT;
	if (info->flags & S16D_BUSY)
	 { *stray = (pending & ~S16D_INTDEV_S);
	   return (pending & S16D_INTDEV_S);
	 }
	else
	 { /* device isn't open; therefore, any interrupt is a stray */
	   *stray = pending;
	   return (0);
	 }
}
