/*---------------------------------------------------------------------------
 *		
 * NAME:	rtpatch - insert a patch around the super-user requirement
 *			  for real-time support in Solaris.
 *
 * USAGE:	rtpatch
 *
 *		Must be run as "root" in order to execute correctly.
 *
 * DESCRIPTION: This program patches a Solaris kernel to remove the 
 *		superuser checks which are normally required to lock
 *		memory and put a process priority into the real-time class.
 *
 * NOTES:	This program works by spawing an 'adb' subprocess.
 *
 * AUTHOR/DATE: John Lloyd, Summer 1996.
 *
 * UPDATES:
 *
 *--------------------------------------------------------------------------*/

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <fcntl.h>
#include <stdarg.h>

/* These includes are just to implement a function which gets the
   RT scheduling class loaded into the kernel */
#include <sys/types.h>
#include <sys/priocntl.h>
#include <sys/rtpriocntl.h>
#include <sys/tspriocntl.h>
#include <sys/iapriocntl.h>
#include <sys/lwp.h>
#include <errno.h>
 
#define ARRAY_SIZE(a)	(sizeof(a)/sizeof(a[0]))   /* # of array elements */

typedef struct {
	char *fxnName;
	char *callName;
	int cnt;
} PatchSpec;

PatchSpec patchList[] = {
	"rt_enterclass", "suser", -1,
	"memcntl", 	 "suser", -1

};

char *extractHexStr (char *s2, const char *s1)
{
	const char *cp1;
	char *cp2;

	cp1 = s1;
	while (isspace (*cp1))
	 { cp1++;
	 }
	if (isxdigit (*cp1))
	 { cp2 = s2;
	   while (isxdigit (*cp1))
	    { *cp2++ = *cp1++;
	    }
	   *cp2 = '\0';
	   return (cp2);
	 }
	else
	 { return (NULL);
	 }
}

#define RBUFSIZE	1024

char rbuf[RBUFSIZE];
char **argvec;

#define SYS	-1

FILE *wIn;		/* input from adb 'write' process */
FILE *wOut;		/* output to adb 'write' process */
FILE *sIn;		/* input from adb 'search' process */
FILE *sOut;		/* output to adb 'search' process */
int verbose = 0;	/* diagnostic output */
int fake = 0;		/* don't do any writing */
int wPid;		/* id for adb 'write' process */
char suserAddr[32];

static fatal (int code, char *fmt, ...);
static output (FILE *fp, char *fmt, ...);
static input();

testc (fp)
FILE *fp;
{
	/* First check to see if there is any data in the buffer */

	if (fp->_cnt > 0)
	 { return (1);
	 }
	else
	 { return (testinput(fileno(fp)));
	 }
}

testinput (fd)
int fd;
{
	struct timeval timeout;		/* timeout variable for select()   */
	fd_set readmask;		/* IO detection mask for select()  */
	int status;			/* status returned by select()	   */
 
	timeout.tv_sec = timeout.tv_usec = 0;

	/* Now perform the test on the stdin stream */
 
	FD_ZERO (&readmask);
	FD_SET (fd, &readmask);
	if (fd >= FD_SETSIZE)
	 { fprintf (stderr,
"%s: testinput() -- fd %d >= FD_SETSIZE\n", *argvec, fd);
	   exit (1);
	 }
	status = select ((1<<fd), &readmask, NULL, NULL, &timeout);
 
	if (status < 0)
	 { fatal (SYS, "testinput: select\n");
	 }
	else
	 { return (FD_ISSET(fd, &readmask));
	 }
}

void onChld(sig)
int sig;
{
	fatal (1, "abd process died\n");
}

loadRTSchedClass()
{
	pcinfo_t info;
	int nc;

	(void) strcpy (info.pc_clname, "RT");
	nc = priocntl (0L, 0L, PC_GETCID, (caddr_t)&info);

	if (nc == -1L)
	 { fatal (SYS, "loadRTSchedClass: priocntl\n");
	 }
}

main(argc, argv)
int argc;
char **argv;
{
	/* The program works by creating two adb processes: 

	   One is used to write the patch into the object file;
	   the other is used to search /vmunix to find the proper
	   location at which to apply the patch.
	 */

	char matchStr[RBUFSIZE];
	int fds1[2];		/* file descriptors */
	int fds2[2];		/*   returned by pipe() */
	int k;

	argvec = argv;

	argc--; argv++;
	while (argc)
	 { if (strcmp (*argv, "-v") == 0)
	    { verbose = 1;
	    }
	   else if (strcmp (*argv, "-N") == 0)
	    { fake = 1;
	    }
	   else
	    { argc = -1;
	      break;
	    }
	   argc--; argv++;
	 }
	if (argc == -1)
	 { fprintf (stderr, 
"usage: %s [-v] [-N]\n", argvec[0]);
	   exit (1);
	 }

	loadRTSchedClass();	

	/* start by trying to create the adb write process, which will
	   actually do the patching: */

	if (pipe(fds1) < 0 || pipe(fds2) < 0)
	 { fatal (SYS, "can't open pipes for adb write process\n");
	 }
	signal (SIGCHLD, onChld);
	if ((wPid = fork()) == 0)
	 { int k;
	   if (dup2(fds1[0], fileno(stdin)) < 0 ||
	       dup2(fds2[1], fileno(stdout)) < 0)
	    { kill (getppid(), SIGKILL);
	      fatal (SYS, "can't map pipes onto stdin and stdout\n");
	    }
	   for (k=32; k>2; k--) close(k);
	   if (execl("/usr/bin/adb", "adb","-w","-k","/dev/ksyms","/dev/mem",0) < 0)
	    { kill (getppid(), SIGKILL);
	      fatal (SYS, "can't exec 1st 'adb' to do patching\n");
	    }
	   /* not reached */
	 }
	/* we are the parent, connected to the child */
	if ((wIn = fdopen(fds2[0], "r")) == NULL ||
	    (wOut = fdopen(fds1[1], "w")) == NULL)
	 { fatal (SYS, "can't do fdopen() on pipe connections\n");
	 }

	/* poke the first adb and check that it's happy */
	output (wOut, "\n");
	input (wIn, rbuf, RBUFSIZE);
 	flushInput (wIn);
	output (wOut, "%s+4?i\n", patchList[0].fxnName);
	sprintf (matchStr, "^%s", patchList[0].fxnName);
	input (wIn, rbuf, RBUFSIZE);
	if (!match (matchStr, rbuf))
	 { fprintf (stderr, "%s", rbuf);
	   fatal (1, "adb write process is not happy\n");
	 }
	output (wOut, "suser=X\n");
	input (wIn, rbuf, RBUFSIZE);
	if (extractHexStr (&suserAddr[1], rbuf) == NULL)
	 { fprintf (stderr, "%s", rbuf);
	   fatal (1, "couldn't find hex address for suser\n");
	 }
	suserAddr[0] = '<';
	strcat (suserAddr, ">");

	for (k=0; k<ARRAY_SIZE(patchList); k++)
	 { doPatch (&patchList[k]);
	 }
}

#define MAXCALLS	16

doPatch (spec)
PatchSpec *spec;
{
	char matchStr[RBUFSIZE];
	char iaddr[MAXCALLS][RBUFSIZE];
	int cnt, k;

	sleep(1);

	output (wOut, "%s?i\n", spec->fxnName);
	sprintf (matchStr, "^%s", spec->fxnName);
	input (wIn, rbuf, RBUFSIZE);
	cnt = 0;
	k = 0;
	while (match(matchStr, rbuf))
	 { if (match("call", rbuf) && match(suserAddr, rbuf))
	    { if (cnt >= MAXCALLS)
	       { fatal (1, "more than %d calls to suser in %s\n",
			      MAXCALLS, spec->fxnName);
	       }
	      strcpy (iaddr[cnt++], rbuf);
	    }
	   output (wOut, "\n");
	   input (wIn, rbuf, RBUFSIZE);
	   k++;
	 }
	flushInput (wIn);
	if (k == 0)
	 { fprintf (stderr, "Warning: symbol %s not found\n", spec->fxnName);
	   return;
	 }
	if (cnt == 0)
	 { fprintf (stderr, "Warning: No calls to suser found in %s\n", 
			     spec->fxnName);
	   return;
	 }
	if (spec->cnt > 0 && k != spec->cnt)
	 { fprintf (stderr, "Warning: %d (vs. %d) calls to suser found in %s\n",
		      k, spec->cnt, spec->fxnName, spec->cnt);
	 }
	for (k=0; k<cnt; k++)
	 { if (!strchr (iaddr[k], ':'))
	    { fatal (1, "address for suser call in %s not colon terminated\n",
			 spec->fxnName);
	    }
	   *strchr(iaddr[k],':') = '\0';
	 }
	for (k=0; k<cnt; k++)
	 { 
	   /* The sequence 

	    	call   +0xea877b76 <e0046e10>
	     	addl   $0x4,%esp [4,-]

	      is replaced with

		movl $0x1,%eax
		addl   $0x4,%esp [4,-]

	      code for movl $0x1,%eax is 00000001b8 (5 bytes)
	      This means replacing 

	      x:   call +0xea877b76 <e0046e10>	with 000001b8
	      x+4: 04c483??			with 04c48300
	    */

	   if (fake)
	    { fprintf (stderr, "W> ");
	      fprintf (stderr, "%s?W 90102001\n", iaddr[k]);
	      fprintf (stderr, "W> ");
	      fprintf (stderr, "&+4?W 1000000\n");
	    }
	   else
	    { 
	      output (wOut, "%s?W 1b8\n", iaddr[k]);
	      input (wIn, rbuf, RBUFSIZE);
	      output (wOut, "&+4?W 4c48300\n");
	      input (wIn, rbuf, RBUFSIZE);

	      /* read back the result and check it */
	      output (wOut, "%s?i\n", iaddr[k]);
	      input (wIn, rbuf, RBUFSIZE);
	      if (!match("movl", rbuf) || !match("$0x1,%eax", rbuf))
	       { fatal (1, 
"call %d to suser in %s: movl $0x1,%eax does not read back\n",
			k+1, spec->fxnName);
	       }
	      output (wOut, "&+5?i\n");
	      input (wIn, rbuf, RBUFSIZE);
	      if (!match("addl", rbuf) || !match("$0x4,%esp", rbuf))
	       { fatal (1, 
"call %d to suser in %s: addl $0x4,%esp does not read back\n",
			k+1, spec->fxnName);
	       }
	    }
	 }
	return 0;
}

static fatal (int code, char *fmt, ...)
{
	/* Terminate the program with suitable error messages. */

	va_list pvar;

	extern char *sys_errlist[];
	extern int errno;

	va_start(pvar, fmt);
	fprintf (stderr, "%s: ", argvec[0]);
	vfprintf (stderr, fmt, pvar);
	va_end(pvar);

	if (code == SYS)
	 { perror ("");
	   code = 1;
	 }
	if (wPid)
	 { kill (wPid, SIGKILL);
	 }
	exit (code);
}

static output (FILE *fp, char *fmt, ...)
{
	/* wrapper for fprintf() */

	va_list pvar;

	va_start(pvar,fmt);
	vfprintf (fp, fmt, pvar);
	fflush (fp);
	if (verbose)
	 { printf ("%s> ", fp == wOut ? "W" : fp == sOut ? "S" : "?" );
	   vfprintf (stdout, fmt, pvar);
	 }
	va_end(pvar);
}

static input(fp, buf, size)
FILE *fp;
char *buf;
int size;
{
	/* wrapper for fgets() */
 
	do
	 { fgets (buf, size, fp);
	   if (verbose)
	    { printf ("%s< ", fp == wIn ? "W" : fp == sIn ? "S" : "?" );
	      printf ("%s", buf);
	    }
	 }
	while (fp->_cnt > 0);
}

match(s1,s2)
char *s1, *s2;
{
	/* return true if string 's1' matches a substring inside 's2'.
	   A '^' character at the beginning of 's1' is used to match
	   the beginning of 's2'.
	 */
	register char *cp;
	register int l1, l2;

	if (*s1 == '^')
	 { return (strncmp(s1+1,s2,strlen(s1+1))==0);
	 }
	else
	 { l1 = strlen(s1);
	   cp = s2;
	   l2 = strlen(s2);
	   while (l2 >= l1)
	    { register char *p1, *p2;
	      p1 = s1;
	      p2 = cp;
	      while (*p1++ == *p2++)
	       { if (*p1 == '\0')
		  { return (1);
		  }
	       }
	      cp++;
	      l2--;
	    }	
	 }
	return(0);
}

flushInput(fp)
FILE *fp;
{
	char buf[1024];
	int flags;
	int nc = 0;

	sleep(1);
	while (testc (fp))
	 { buf[nc++] = getc(fp);
	 }
	if (verbose)
	 { printf ("flushInput read %d chars:\n", nc);
	 }
	if (nc > 0 && verbose)
	 { printf ("%s< ", fp == wIn ? "W" : fp == sIn ? "S" : "?" );
	   buf[nc] = '\0';
	   printf ("%s", buf);
	   fflush (stdout);
	 }
}
