#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/sysmacros.h>
#include <sys/stat.h>

#include "atiioctl.h"
#include "atireg.h"
#include <enableIO.h>

#define STATIC
#define MAPPED_IO	0

int atidevflag = 0;
void *atii;

#define ATI_INTR_SPEC	0	/* device interrupt spec. */
#define ATI_REG_SPEC	0	/* register set to be mapped */
#define ATI_NULL_HANDLER (u_int(*)(caddr_t))nulldev

/* maximum number of instances supported by this driver */
#define MAX_INSTANCES		4
/* number of bits in the minor node reserved for the instance number */
#define NUM_INSTANCE_BITS	4
/* returns the instance number from a dev_t structure */
#define GET_INSTANCE(d)		getminor(d)

/* Hardware-specific device routines */
static int atiHardwareReset (AtiSoftc*);
static int atiHardwareOpen (AtiSoftc*);
static int atiHardwareClose (AtiSoftc* );
static int atiHardwareIntrEnable (AtiSoftc*, int);
static int atiHardwareIntrHandle (AtiSoftc*, int*);

static void atiReadData (AtiForceRec *rec, AtiSoftc *info);
static int atiReadCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info);
static int atiWriteCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info);

/* Routines to access the force data buffer */

static int fbufRecsFree (AtiSoftc *info);
static int fbufRecsQueued (AtiSoftc *info);
static void fbufAdvanceTail (AtiSoftc *info);
static void fbufAdvanceHead (AtiSoftc *info);
static int fbufNextTail (AtiSoftc *info);
static int fbufNextHead (AtiSoftc *info);

/* Stuff relating to dynamic loading and autoconfiguration. There
   is *a lot* of boiler plate associated with this. */

#define ATI_SPIN_TIMEOUT	10000

static int atiidentify();
static int atiattach();
static int atidetach();
static int atigetinfo();

static int atiopen(dev_t *devp, int flag, int otyp, cred_t *credp);
static int aticlose(dev_t dev, int flag, int otyp, cred_t *credp);
static int atiread(dev_t dev, struct uio *uiop, cred_t *credp);
static int atiwrite(dev_t dev, struct uio *uiop, cred_t *credp);
static int atiioctl(dev_t dev, int cmd, int arg, int mode, 
			cred_t *credp, int *rvalp);
static int atimmap(dev_t dev, off_t off, int prot);
static int atichpoll(dev_t dev, short events, int anyyet,
		    short *reventsp, struct pollhead **phpp);

static u_int atiintr (caddr_t);
static u_int atisoftintr (caddr_t);

static struct cb_ops ati_cb_ops = {
	atiopen,
	aticlose,
	nodev,
	nodev,
	nodev,
	atiread,
	atiwrite,
	atiioctl,
	nodev,
	atimmap,
	nodev,
	atichpoll,
	nodev,
	NULL,
	D_NEW | D_MP
};

static struct dev_ops ati_ops = {
	DEVO_REV,
	0,
	atigetinfo,
	atiidentify,
	nulldev,
	atiattach,
	atidetach,
	nodev,
	&ati_cb_ops,
	(struct bus_ops *)NULL
};

static struct modldrv modldrv = {
	&mod_driverops,
	"ati driver",
	&ati_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	NULL,
};

int _init (void)
{
	int status;

	ddi_soft_state_init (&atii, sizeof(AtiSoftc), MAX_INSTANCES);
	status = mod_install (&modlinkage);
	if (status != 0)
	 { ddi_soft_state_fini (&atii);
	 }
	return (status);
}

int _fini (void)
{
	int status;

	status = mod_remove (&modlinkage);
	if (status == 0)
	 { ddi_soft_state_fini (&atii);
	 }
	return (status);
}

int _info (struct modinfo *modinfop)
{
	return (mod_info (&modlinkage, modinfop));
}

static int atiidentify (dip)
dev_info_t *dip;
{
	/* For SBus devices, the name checked by this routine must match
	   the name stored on the device itself. */

	if (strcmp (ddi_get_name(dip), "ati") == 0)
	 { return (DDI_IDENTIFIED);
	 }
	else
	 { return (DDI_NOT_IDENTIFIED);
	 }
}

static int atigetinfo (dev_info_t *dip, ddi_info_cmd_t infocmd, 
			  void *arg, void **result)
{
	AtiSoftc *info;
	int status;

	switch (infocmd)
	 { case DDI_INFO_DEVT2INSTANCE:
	    { *result = (void*)GET_INSTANCE((dev_t)arg);
	      status = DDI_SUCCESS;
	      break;
	    }
	   case DDI_INFO_DEVT2DEVINFO:
	    { info = ddi_get_soft_state(atii, GET_INSTANCE((dev_t)arg));
	      if (info == NULL)
	       { return (DDI_FAILURE);
	       }
	      *result = (void*)info->dip;
	      status = DDI_SUCCESS;
	      break;
	    }
	   default:
	    { status = DDI_FAILURE;
	      break;
	    }
	 }
	return (status);
}

static void atifreeResources (dev_info_t *dip, AtiSoftc *info)
{
	if (info->fsize)
	 { kmem_free (info->fbuf, sizeof(AtiForceRec)*info->fsize);
	 }
	if (info->resources & ATI_MAIN_MINOR_CREATED)
	 { ddi_remove_minor_node (dip, NULL);
	 }
	if (info->resources & ATI_IHANDLER_SET)
	 { ddi_remove_intr (dip, ATI_INTR_SPEC, &info->iblkCookie);
	 }
	if (info->resources & ATI_SOFT_IHANDLER_SET)
	 { ddi_remove_softintr (info->softIntrId);
	 }
	if (info->resources & ATI_LOW_MUTEX_SET)
	 { mutex_destroy (&info->lowMux);
	 }
	if (info->resources & ATI_INTR_MUTEX_SET)
	 { mutex_destroy (&info->intrMux);
	 }
	if (info->resources & ATI_REGS_MAPPED)
	 { ddi_regs_map_free (&info->regsHandle);
	 }
	ddi_soft_state_free (atii, info->instance);
}

static int atiattach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	AtiSoftc *info;
	int instance;
	ddi_device_acc_attr_t dev_attr;

	static char *mainNames[] = { "0", "1", "2", "3" };

	switch (cmd)
	 { case DDI_ATTACH:
	    { instance = ddi_get_instance(dip);
	      if (instance >= MAX_INSTANCES)
	       { return (DDI_FAILURE);
	       }
	      if (ddi_soft_state_zalloc(atii, instance) != 0)
	       { return (DDI_FAILURE);
	       }
	      info = ddi_get_soft_state(atii, instance);

	      info->dip = dip;
	      info->instance = instance;
	      enioInit (&info->enio);	
#if MAPPED_IO
	      /* map device registers */
	      if (ddi_map_regs (dip, ATI_REG_SPEC, (caddr_t*)&info->regs,
				/*offset=*/0, /*len=*/0) != DDI_SUCCESS)
	       { atifreeResources (dip, info);
		 cmn_err (CE_CONT, "ati: couldn't map regs\n");
		 return (DDI_FAILURE);
	       }
	      info->resources |= ATI_REGS_MAPPED;
#else
	      dev_attr.devacc_attr_version = DDI_DEVICE_ATTR_V0;
	      dev_attr.devacc_attr_endian_flags = DDI_NEVERSWAP_ACC;
	      dev_attr.devacc_attr_dataorder = DDI_STRICTORDER_ACC;
	      if (ddi_regs_map_setup (dip, ATI_REG_SPEC, 
		  (caddr_t*)&info->regs, /*offset=*/0, /*len=*/0,
		   &dev_attr, &info->regsHandle) == DDI_SUCCESS)
	       { cmn_err (CE_CONT, "ati: regs_map_setup worked: 0x%x\n", 
			  info->regs);
		 info->resources |= ATI_REGS_MAPPED;
	       }
	      else
	       { cmn_err (CE_CONT, "ati: regs_map_setup failed\n");
		 info->regs = 0;
		 atifreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->iobase = 0x280;	/* XXX XXX XXX */
#endif
	      /* make device quiescent */
	      if (atiHardwareReset (info) <0)
	       { cmn_err (CE_CONT, "ati: hardware reset failed\n");
		 atifreeResources (dip, info);
		 return (DDI_FAILURE);
	       }

	      /* set up interrupt-level mutex for the device. Start
		 by assigning a dummy interrupt handler so the
		 mutex is associated with the proper interrupt context,
		 without risk of an a handler getting called. */

	      if (ddi_add_intr (dip, ATI_INTR_SPEC,&info->iblkCookie,
		     NULL, ATI_NULL_HANDLER, NULL) != DDI_SUCCESS)
	       { atifreeResources (dip, info);
		 return (DDI_FAILURE);
	       }

	      mutex_init (&info->intrMux, "ati intr mutex", MUTEX_DRIVER,
			  (void*)info->iblkCookie);
	      ddi_remove_intr (dip, ATI_INTR_SPEC, info->iblkCookie);
	      info->resources |= ATI_INTR_MUTEX_SET;

	      /* see if the device interrupt is delivered in a high-level
		 interrupt context. If it is, we need to set up a 
		 low-level interrupt handler and mutex as well,
		 because we can't call cmn_err() or pollwakeup()
		 from a high-level interrupt context. */

	      if (ddi_intr_hilevel (dip, ATI_INTR_SPEC))
	       { info->hasHighLevelIntr = 1;
		 
		 /* set up low-level mutex and handler */

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			ATI_NULL_HANDLER, NULL) != DDI_SUCCESS)
		  { atifreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 mutex_init (&info->lowMux, "ati low mutex", MUTEX_DRIVER,
			     (void*)info->softIblkCookie);
		 ddi_remove_softintr (info->softIntrId);
		 info->resources |= ATI_LOW_MUTEX_SET;

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			atisoftintr, (caddr_t)info) != DDI_SUCCESS)
		  { atifreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 info->resources |= ATI_SOFT_IHANDLER_SET;

		 info->muxp = &info->lowMux;
		 cmn_err (CE_CONT, "ati: interrupts are high level\n");
	       }
	      else
	       { info->hasHighLevelIntr = 0;
		 info->muxp = &info->intrMux;
		 cmn_err (CE_CONT, "ati: interrupts are low level\n");
	       }

	      if (ddi_add_intr (dip, ATI_INTR_SPEC, &info->iblkCookie,
		     &info->idevCookie, atiintr,(caddr_t)info) != DDI_SUCCESS)
	       { atifreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= ATI_IHANDLER_SET;

	      if (ddi_create_minor_node (dip, mainNames[instance], S_IFCHR, 
		     instance, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { atifreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= ATI_MAIN_MINOR_CREATED;

	      info->fsize = 0;
	      info->fbuf = NULL;
	      info->fhead = info->ftail = 0;
	      info->foverflow = 0;

	      ddi_report_dev(dip);
cmn_err (CE_CONT, "succeeded\n");
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }
}

static int atidetach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	AtiSoftc *info;
	
	switch (cmd)
	 { case DDI_DETACH:
	    { info = ddi_get_soft_state (atii, ddi_get_instance(dip));

cmn_err (CE_CONT, "ati: detach called\n");
	      /* make device quiescent */
	      atiHardwareReset (info);
	      atifreeResources (dip, info);
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }	
}

static u_int atiintr (caddr_t arg)
{
	/* Called in response to an interrupt from the ATI */

	AtiSoftc *info = (AtiSoftc*)arg;
	int needSoftIntr = 0;
	int pending;
	int stray;

	mutex_enter (&info->intrMux);
	pending = atiHardwareIntrHandle (info, &stray);
	if (pending == 0 && stray == 0)
	 { mutex_exit (&info->intrMux);
	   return (DDI_INTR_UNCLAIMED);
	 }

	if (pending)
	 { if (info->hasHighLevelIntr)
	    { if (!info->softIntrRunning)
	       { needSoftIntr = 1;
	       }
	    }
	   else
	    { if (fbufRecsQueued(info) > 0)
	       { mutex_exit (&info->intrMux);
		 pollwakeup (&info->pollhead, POLLRDNORM);
		 mutex_enter (&info->intrMux);
	       }
	    }
	 }
	if (stray)
	 { /* declare a stray interrupt */
	   if (info->hasHighLevelIntr)
	    { info->stray = stray;
	      if (!info->softIntrRunning)
	       { needSoftIntr = 1;
	       }
	    }
	   else
	    { cmn_err (CE_CONT, "ati%d: stray interrupt(s); mask=0x%x\n",
		   		 info->instance, stray);
	    }
	 }
	mutex_exit (&info->intrMux);
	if (needSoftIntr)
	 { ddi_trigger_softintr (info->softIntrId);
	 }
	return (DDI_INTR_CLAIMED);
}

static u_int atisoftintr (caddr_t arg)
{
	/* Used only when the device interrupts occur at high interrupt
	   priority. Calls pollwakeup and cmn-err because the high-level
	   interrupt handler can't. */

	AtiSoftc *info = (AtiSoftc*)arg;
	int stray;

	mutex_enter (&info->lowMux);
	mutex_enter (&info->intrMux);

	/* don't know why info->softIntrRunning needs to be checked
	   in the next phrase, but it was in the manual example */
	if ((info->stray == 0 && fbufRecsQueued(info) == 0) ||
	    info->softIntrRunning)
	 { mutex_exit (&info->intrMux);
	   mutex_exit (&info->lowMux);
	   return (DDI_INTR_UNCLAIMED);	
	 }
	info->softIntrRunning = 1;

	if ((stray = info->stray) != 0)
	 { info->stray = 0;
	 }
	mutex_exit (&info->intrMux);

	if (stray)
	 { cmn_err (CE_CONT, "ati%d: stray interrupt(s); mask=0x%x\n",
			      info->instance, stray);
	 }
	if (fbufRecsQueued(info) > 0)
	 { mutex_exit (&info->lowMux);
	   pollwakeup (&info->pollhead, POLLRDNORM);
	   mutex_enter (&info->lowMux);
	 }
	info->softIntrRunning = 0;
	mutex_exit (&info->lowMux);
	return (DDI_INTR_CLAIMED);	
}

STATIC int atiopen (dev_t *dev, int flag, int otyp, cred_t *crp)
{
	/* This device is an exclusive open device: only one process
	   may open it at a time. This prevents contention problems
	   and makes the driver software easier to implement. */

	AtiSoftc *info;

	if ((info = ddi_get_soft_state(atii,GET_INSTANCE(*dev))) == NULL)
	 { return (ENXIO);
	 }
	if (otyp != OTYP_CHR)
	 { return (EINVAL);
	 }
#if MAPPED_IO
	if (!info->regs)	/* paranoid; this check shouldn't be needed */
	 { return (ENXIO);
	 }
#endif
	mutex_enter (info->muxp);
	if (!(info->flags & ATI_OPEN))
	 { info->ftail = info->fhead = 0;	/* clear fbuffer */
	   info->foverflow = 0;
	   if (atiHardwareOpen (info) <0)
	    { cmn_err (CE_CONT, "ati: hardware open failed\n");
	      return (ENXIO);
	    }
	   info->flags |= ATI_OPEN;
	 }
	mutex_exit (info->muxp);
	return (0);
}

STATIC int aticlose (dev_t dev, int flag, int otyp, cred_t *crp)
{
	AtiSoftc *info;

	if ((info = ddi_get_soft_state(atii,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	mutex_enter (info->muxp);
	if (atiHardwareClose (info) < 0)
	 { cmn_err (CE_CONT, "ati: hardware close failed\n");
	 }
	info->flags = 0;
	mutex_exit (info->muxp);
	return (0);
}

STATIC int atiread (dev_t dev, uio_t *uiop, cred_t *crp)
{
	/* The read routine returns whatever force/torque data 
	   may have been queued */

	AtiSoftc *info;
	int nxfer;	/* number of records to transfer */
	int nhend;	/* number of records from head to end of buffer */
	int nread;	/* number of read buffers available */
	int nextHead;

	if ((info = ddi_get_soft_state(atii,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	nread = uiop->uio_resid/sizeof(AtiForceRec);
	if (nread && info->foverflow)
	 { AtiForceRec rec;

	   info->foverflow = 0;
	   rec.dword = ATI_OVERFLOW;
	   if (uiomove ((caddr_t)&rec,
			sizeof(AtiForceRec),
			UIO_READ, uiop) != 0)
	    { return (EFAULT);
	    }
	   nread--;
	 }
	nxfer = min(nread,fbufRecsQueued(info));
	nextHead = fbufNextHead (info);
	nhend = info->fsize-nextHead;
	if (nxfer)
	 { if (uiomove ((caddr_t)&info->fbuf[nextHead], 
			min(nxfer,nhend)*sizeof(AtiForceRec),
	   		UIO_READ, uiop) != 0)
	    { return (EFAULT);
	    }
	 }
	if (nxfer-nhend > 0)
	 { if (uiomove ((caddr_t)&info->fbuf[0], 
			(nxfer-nhend)*sizeof(AtiForceRec),
	   		UIO_READ, uiop) != 0)
	    { return (EFAULT);
	    }
	 }
	info->fhead = (info->fhead+nxfer)%info->fsize;
	return (0);
}

STATIC int atiwrite (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STATIC int atimmap (dev_t dev, off_t off, int prot)
{
	AtiSoftc *info;

	if ((info = ddi_get_soft_state(atii,GET_INSTANCE(dev))) == NULL)
	 { return (-1);
	 }
	if (off < 0 || off >= sizeof(AtiRegs))
	 { return (-1);
	 }
#if MAPPED_IO
	return (hat_getkpfnum((char*)info->regs + off));
#else
	return (-1);
#endif
}

STATIC int atichpoll (dev_t dev, short events, int anyyet,
		      short *reventsp, struct pollhead **phpp)
{
	/* Check to see if any force data is in the buffer. If there
	   isn't, then queue the process for a possible select wakeup. */

	AtiSoftc *info;
	int nrecs;

	if ((info = ddi_get_soft_state(atii,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	nrecs = fbufRecsQueued(info);

	/* not really sure how system prevents races with the interrupt
	   handler here, but this is what the example says to do ... */
	if ((events & (POLLIN|POLLRDNORM)) && nrecs > 0)
	 { *reventsp = events & (POLLIN|POLLRDNORM);
	 }
	else
	 { *reventsp = 0;
	   if (!anyyet)
	    { *phpp = &info->pollhead;
	    }
	 }
	return (0);
}


STATIC int atiioctl (dev_t dev, int cmd, int arg, int mode, 
		     cred_t *crp, int *rvalp)
{
	AtiSoftc *info;
	int status;

	if ((info = ddi_get_soft_state(atii,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case ATIIOC_GET_SENSOR_INFO:
	    { AtiSensorRec rec;

	      mutex_enter (info->muxp);
	      status = atiReadCom (ATI_SENSOR_DATA_CMD, ATI_READ, &rec,
				   sizeof(rec)/2, info);
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg, sizeof(rec)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_GET_BIAS:
	    { AtiBiasRec rec;

	      mutex_enter (info->muxp);
	      status = atiReadCom (ATI_BIAS_ARRAY_CMD, ATI_READ, &rec,
				   sizeof(rec)/2, info);
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg, sizeof(rec)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_SET_BIAS:
	    { AtiBiasRec rec;
	      
	      if (copyin((caddr_t)arg, (caddr_t)&rec, sizeof(rec)))
	       { return (EFAULT);
	       }
	      mutex_enter (info->muxp);
	      status = atiWriteCom (ATI_BIAS_ARRAY_CMD, 2, &rec,
				    sizeof(rec)/2, info);
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      break;
	    }
	   case ATIIOC_RESET_BIAS:
	    { mutex_enter (info->muxp);
	      status = atiWriteCom (ATI_BIAS_ARRAY_CMD, 1, NULL, 0, info);
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      break;
	    }
	   case ATIIOC_TEST_RESET:
	    { AtiTestRec rec;
	      
	      mutex_enter (info->muxp);
	      status = atiReadCom (ATI_RESET_CMD, ATI_NO_MODE, &rec,
				   sizeof(rec)/2, info);
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg, sizeof(rec)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_SET_CAPBUF_SIZE:
	    { if (arg <= 0)
	       { return (EINVAL);
	       }
	      mutex_enter (info->muxp);
	      if (info->fsize != arg+1)
	       { if (info->fsize)
		  { kmem_free (info->fbuf, info->fsize*sizeof(AtiForceRec));
		    info->fsize = 0;
		  }
		 info->fbuf = (AtiForceRec*)kmem_alloc (
			       (arg+1)*sizeof(AtiForceRec), KM_NOSLEEP);
		 if (info->fbuf == NULL)
		  { mutex_exit (info->muxp);
		    return (ENOMEM);
		  }
		 info->fsize = (arg+1);
	       }
	      mutex_exit (info->muxp);
	      break;
	    }
	   case ATIIOC_GET_CAPBUF_SIZE:
	    { int size = info->fsize-1;
	      if (size < 0)
	       { size = 0;
	       }
	      if (copyout((caddr_t)&size, (caddr_t)arg, sizeof(int)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_GET_FORCE_DATA:
	    { AtiForceRec rec;

	      mutex_enter (info->muxp);
	      atiReadData (&rec, info);
	      mutex_exit (info->muxp);
	      if (copyout((caddr_t)&rec, (caddr_t)arg, sizeof(rec)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_SET_CALMATRIX:
	    { AtiCalMatrixRec rec;

	      if (copyin((caddr_t)arg, (caddr_t)&rec, sizeof(rec)))
	       { return (EFAULT);
	       }
	      mutex_enter (info->muxp);
	      status = atiWriteCom (ATI_CALMATRIX_CMD, ATI_WRITE, &rec,
				    sizeof(rec)/2, info);
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      break;
	    }
	   case ATIIOC_GET_CALMATRIX:
	    { AtiCalMatrixRec rec;

	      mutex_enter (info->muxp);
	      status = atiReadCom (ATI_CALMATRIX_CMD, ATI_READ, &rec,
				   sizeof(rec)/2, info);
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg,sizeof(rec)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_GET_PEAK_DATA:
	    { AtiPeakValueRec rec;

	      mutex_enter (info->muxp);
	      status = atiReadCom (ATI_GET_PEAK_CMD, ATI_READ, &rec,
				   sizeof(rec)/2, info);
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg,sizeof(rec)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_GET_PEAK_RESET:
	    { AtiPeakValueRec rec;

	      /* ATI_WRITE here means to reset the data after returning it */
	      mutex_enter (info->muxp);
	      status = atiReadCom (ATI_GET_PEAK_CMD, ATI_WRITE, &rec,
				   sizeof(rec)/2, info);
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg,sizeof(rec)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_GET_OPFLAGS:
	    { int oflags;

	      oflags = info->cdata.oflags;
	      if (copyout((caddr_t)&oflags, (caddr_t)arg,sizeof(int)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_SET_OPFLAGS:
	    { unsigned short oflags = (unsigned short)arg;
	    
	      mutex_enter (info->muxp);
	      if (oflags != info->cdata.oflags)
	       { info->cdata.oflags = oflags;
		 status = atiWriteCom (ATI_CARD_DATA_CMD, ATI_WRITE, &info->cdata,
				      sizeof(info->cdata)/2, info);
	       }
	      else
	       { status = 0;
	       }
	      mutex_exit (info->muxp);
	      if (status < 0)
	       { return (ETIME);
	       }
	      break;
	    }
	   case ATIIOC_GET_CAPTURE_RATE:
	    { int rate;

	      mutex_enter (info->muxp);
	      if (info->flags & ATI_CAPTURE_ON)
	       { rate = info->cdata.sampleRate;
	       }
	      else
	       { rate = 0;
	       }
	      mutex_exit (info->muxp);
	      if (copyout((caddr_t)&rate, (caddr_t)arg,sizeof(int)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case ATIIOC_SET_CAPTURE:
	    {
	      if (arg > 4000)
	       { return (EINVAL);
	       }
	      if (arg <= 0)
	       { arg = 0;
	       }
	      mutex_enter (info->muxp);
	      info->cdata.sampleRate = arg;
	      if (info->fsize == 0)
	       { info->fbuf = (AtiForceRec*)kmem_alloc (
		       (ATI_DEFAULT_FBUF_SIZE+1)*sizeof(AtiForceRec), 
			KM_NOSLEEP);
		 if (info->fbuf == NULL)
		  { mutex_exit (info->muxp);
		    return (ENOMEM);
		  }
		 info->fsize = (ATI_DEFAULT_FBUF_SIZE+1);
	       }
	      if (info->cdata.intrEnable == 0)
	       { info->fhead = 0;
		 info->ftail = 0;
		 info->foverflow = 0;
	       }	      
	      if (atiHardwareIntrEnable (info, arg != 0) < 0)
	       { mutex_exit (info->muxp);
		 return (ETIME);
	       }
	      info->flags |= ATI_CAPTURE_ON;
	      mutex_exit (info->muxp);
	      break;
	    }
	   case ATIIOC_ENABLE_IOPORTS:
	    {
	      if (arg)
	       { enioCall0 (&info->enio);
	       }
	      else
	       { enioCall1 (&info->enio);
	       }
	      break;
	    }
	 }
	return (0);
}


/*------------------------ */

#if MAPPED_IO
#define OUTB(p,v)	regs->p = (v)
#define OUTW(p,v)	regs->p = (v)
#define INB(p)		regs->p
#define INW(p)		regs->p
#else
#define OUTB(p,v)	outb(info->iobase+ATI_##p,v)
#define OUTW(p,v)	outw(info->iobase+ATI_##p,v)
#define INB(p)		inb(info->iobase+ATI_##p)
#define INW(p)		inw(info->iobase+ATI_##p)
#endif

static void atiReadData (AtiForceRec *rec, AtiSoftc *info)
{
#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	int i;

	OUTW (addr, 0x101);
	rec->dword = INW (data);
	if (rec->dword)
	 { OUTW (addr, 0x101);
	   OUTW (data, 0);	/* clear monitor condition */
	 }
	for (i=0; i<6; i++)
	 { rec->fdata[i] = INW (data);
	 }
}

static int atiReadCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info)
{
#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	unsigned short *dbuf = (unsigned short *)ptr;
	int retcode;

	if (atiWriteCom (com, mode, NULL, 0, info) < 0)
	 { return (-1);
	 }
	OUTW(addr,0);	  /* reset address */
	retcode = INW(data);
	if (mode != ATI_NO_MODE)
	 { INW(data);
	 }
	while (n--)
	 { *dbuf++ = INW(data);
	 }
	return (retcode);
}

static int atiWriteCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info)
{
#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	unsigned short *dbuf = (unsigned short *)ptr;
	unsigned dword, retcode;
	int spin;

	/*
	   write command block to DPRAM at offset 0x000 
	 */
	OUTW(addr,0); 		/* set address register to command_block */
	OUTW(data,com);
	if (mode != ATI_NO_MODE)
	 { OUTW(data,mode);
	 }
	if (dbuf)
	 { while (n--)
	    { OUTW(data,*dbuf++);
	    }
	 }
	/*
	   read MailBox at DPRAM 0x3FE to clear host interrupt flag
	 */
	OUTW(addr,0x3FE);	     	/* set mailbox address */
	INW(data);		   	/* reset interrupt flag */
	OUTW(addr,0x3FE) ;	    	/* set mailbox address */
	OUTW(data,com);		    	/* clear acknowledgment */

	/*
	   write command word to Mailbox at DPRAM 0x3FF to set DSP intr flag
	 */
	OUTW(addr,0x3ff) ;
	OUTW(data,com) ;

	/*
	   waits for command completion (interrupt flag) from DSP
	 */
	retcode = 1;
	for (spin = ATI_SPIN_TIMEOUT; spin; spin--)
	 {
	   dword = INW(status);
	   // wait for interrupt flag
	   if((dword & 0x0010)==0)     		/* interrupt set */
	    {
	      /* read mailbox location at 0x3FE to reset interrupt flag */
	      OUTW(addr,0x3FE) ;	 	/* write dpram address */
	      retcode = INW(data) ;	 	/* reset interrupt flag */

	      /* test mailbox for command acknowledgment */
	      if(retcode & 0x00ff)		/* test for ack */
	       { break;
	       }
	    }
	 }
	if (spin)
	 { return retcode;
	 }
	else
	 { return -1;
	 }
}

static int atiHardwareReset (AtiSoftc *info)
{
	/* make the device quiescent. Called at attach and detach time */

	/* get card info */
	if (atiReadCom (ATI_CARD_DATA_CMD, ATI_READ, &info->cdata,
			sizeof(info->cdata)/2, info) < 0)
	 { return (-1);
	 }
	return (atiHardwareIntrEnable (info, 0));
}

static int atiHardwareClose(AtiSoftc *info)
{
	/* make the device quiet when it is closed */

	/* disable interrupts */
	return (atiHardwareIntrEnable (info, 0));
}

static int atiHardwareOpen (AtiSoftc *info)
{
	/* initialize the device when it is openned */	

	if (atiReadCom (ATI_CARD_DATA_CMD, ATI_READ, &info->cdata,
			sizeof(info->cdata)/2, info) < 0)
	 { return (-1);
	 }
	return (0);
}

static int atiHardwareIntrEnable (AtiSoftc *info, int enable)
{
	/* enable/disable device interrupts */

	if (enable != info->cdata.intrEnable)
	 { info->cdata.intrEnable = enable;
	   if (atiWriteCom (ATI_CARD_DATA_CMD, ATI_WRITE, &info->cdata,
			    sizeof(info->cdata)/2, info) < 0)
	    { return (-1);
	    }
	 }
	return (1);
}

static int atiHardwareIntrHandle (AtiSoftc *info, int *stray)
{
	/* clear interrupts, return true if a valid interrupt is pending,
	   and return a mask of any stray interrupts in *stray. */

#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	unsigned dword;

	dword = INW(status);
	if (dword)
	 { OUTW (status, 1);
	   if (info->flags & ATI_CAPTURE_ON)
	    { *stray = 0;
	      if (fbufRecsFree (info) > 0)
	       { int nextTail;
		 nextTail = fbufNextTail(info);
		 atiReadData (&info->fbuf[nextTail], info);
		 info->ftail = nextTail;
	       }
	      else
	       { info->foverflow = 1;
	       }
	      return (1);
	    }
	   else
	    { *stray = 1;
	      return (0);
	    }
	 }
	else
	 { *stray = 0;
	 }
	return (0);
}

static int fbufRecsFree (AtiSoftc *info)
{
	/* returns number of records left in the force data buffer */
	int n;

	if ((n = info->ftail-info->fhead) < 0)
	 { return (-n-1);
	 }
	else
	 { return (info->fsize-n-1);
	 }
}

static int fbufRecsQueued (AtiSoftc *info)
{
	int n;
 
	if ((n = info->ftail-info->fhead) < 0)
	 {  n += info->fsize;
	 }
	return (n);
}

static void fbufAdvanceTail (AtiSoftc *info)
{
	/* advance tail by 1 */
	info->ftail = ((info->ftail+1)%info->fsize);
}

static void fbufAdvanceHead (AtiSoftc *info)
{
	/* advance head by 1 */
	info->fhead = ((info->fhead+1)%info->fsize);
}

static int fbufNextTail (AtiSoftc *info)
{
	/* get advanced tail value */
	return ((info->ftail+1)%info->fsize);
}

static int fbufNextHead (AtiSoftc *info)
{
	/* get advanced head value */
	return ((info->fhead+1)%info->fsize);
}
