#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/sysmacros.h>
#include <sys/stat.h>

#include <pmcioctl.h>
#include <pmcreg.h>
#include <enableIO.h>

#define STPMCC

int pmcdevflag = 0;
void *pmci;

#define PMC_INTR_SPEC	0	/* device interrupt spec. */
#define PMC_REG_SPEC	0	/* register set to be mapped */
#define PMC_NULL_HANDLER (u_int(*)(caddr_t))nulldev

/* maximum number of instances supported by this driver */
#define MAX_INSTANCES		4
/* number of bits in the minor node reserved for the instance number */
#define NUM_INSTANCE_BITS	4
/* returns the instance number from a dev_t structure */
#define GET_INSTANCE(d)		(getminor(d) & ((1<<NUM_INSTANCE_BITS)-1))
/* returns channel device or -1 if device is not a channel device */
#define GET_CHAN_DEV(d)		((getminor(d) >> NUM_INSTANCE_BITS)-1)
/* returns the minor code associated with a channel device */
#define CHAN_MINOR(n)		(((n)+1) << NUM_INSTANCE_BITS)

/* Hardware-specific device routines */
static int pmcHardwareProbe (int base);
static int pmcHardwareReset (PmcSoftc*);
static int pmcHardwareOpen (PmcSoftc*);
static int pmcHardwareClose (PmcSoftc* );
static int pmcHardwareIntrEnable (PmcSoftc*, int);
static int pmcHardwareIntrHandle (PmcSoftc*, int*);
static void pmcHardwareSetPit (PmcSoftc*, int);

/* Hardware-specific utility routines */
static unsigned short pmcReadReg (int iobase, int reg);
static void pmcReadRegs (int iobase, unsigned short *buf, int n, int reg);
static unsigned long pmcReadDSPReg (int iobase, int reg);
static void pmcWriteReg (int iobase, int reg, unsigned short val);
static void pmcWriteRegs (int iobase, int reg, unsigned short *buf, int n);
static void pmcWriteDSPReg (int iobase, int reg, unsigned long val);

/* Stuff relpmcng to dynamic loading and autoconfigurpmcon. There
   is *a lot* of boiler plate associated with this. */

#define PMC_SPIN_TIMEOUT	10000

static int pmcidentify();
static int pmcattach();
static int pmcdetach();
static int pmcgetinfo();
static int pmcprobe();

static int pmcopen(dev_t *devp, int flag, int otyp, cred_t *credp);
static int pmcclose(dev_t dev, int flag, int otyp, cred_t *credp);
static int pmcread(dev_t dev, struct uio *uiop, cred_t *credp);
static int pmcwrite(dev_t dev, struct uio *uiop, cred_t *credp);
static int pmcioctl(dev_t dev, int cmd, int arg, int mode, 
			cred_t *credp, int *rvalp);
static int pmcmmap(dev_t dev, off_t off, int prot);
static int pmcchpoll(dev_t dev, short events, int anyyet,
		    short *reventsp, struct pollhead **phpp);

static u_int pmcintr (caddr_t);
static u_int pmcsoftintr (caddr_t);

static struct cb_ops pmc_cb_ops = {
	pmcopen,
	pmcclose,
	nodev,
	nodev,
	nodev,
	pmcread,
	pmcwrite,
	pmcioctl,
	nodev,
	pmcmmap,
	nodev,
	pmcchpoll,
	nodev,
	NULL,
	D_NEW | D_MP
};

static struct dev_ops pmc_ops = {
	DEVO_REV,
	0,
	pmcgetinfo,
	pmcidentify,
	pmcprobe,
	pmcattach,
	pmcdetach,
	nodev,
	&pmc_cb_ops,
	(struct bus_ops *)NULL
};

static struct modldrv modldrv = {
	&mod_driverops,
	"pmc driver",
	&pmc_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	{ &modldrv, NULL }
};

int _init (void)
{
	int status;

	ddi_soft_state_init (&pmci, sizeof(PmcSoftc), MAX_INSTANCES);
	status = mod_install (&modlinkage);
	if (status != 0)
	 { ddi_soft_state_fini (&pmci);
	 }
	return (status);
}

int _fini (void)
{
	int status;

	status = mod_remove (&modlinkage);
	if (status == 0)
	 { ddi_soft_state_fini (&pmci);
	 }
	return (status);
}

int _info (struct modinfo *modinfop)
{
	return (mod_info (&modlinkage, modinfop));
}

static int pmcidentify (dip)
dev_info_t *dip;
{
	/* For SBus devices, the name checked by this routine must match
	   the name stored on the device itself. */

	if (strcmp (ddi_get_name(dip), "pmc") == 0)
	 { return (DDI_IDENTIFIED);
	 }
	else
	 { return (DDI_NOT_IDENTIFIED);
	 }
}

static int pmcgetinfo (dev_info_t *dip, ddi_info_cmd_t infocmd, 
			  void *arg, void **result)
{
	PmcSoftc *info;
	int status;

	switch (infocmd)
	 { case DDI_INFO_DEVT2INSTANCE:
	    { *result = (void*)GET_INSTANCE((dev_t)arg);
	      status = DDI_SUCCESS;
	      break;
	    }
	   case DDI_INFO_DEVT2DEVINFO:
	    { info = ddi_get_soft_state(pmci, GET_INSTANCE((dev_t)arg));
	      if (info == NULL)
	       { return (DDI_FAILURE);
	       }
	      *result = (void*)info->dip;
	      status = DDI_SUCCESS;
	      break;
	    }
	   default:
	    { status = DDI_FAILURE;
	      break;
	    }
	 }
	return (status);
}

static void pmcfreeResources (dev_info_t *dip, PmcSoftc *info)
{
	if (info->resources & PMC_MINOR_CREATED)
	 { ddi_remove_minor_node (dip, NULL);
	 }
	if (info->resources & PMC_IHANDLER_SET)
	 { ddi_remove_intr (dip, PMC_INTR_SPEC, &info->iblkCookie);
	 }
	if (info->resources & PMC_SOFT_IHANDLER_SET)
	 { ddi_remove_softintr (info->softIntrId);
	 }
	if (info->resources & PMC_LOW_MUTEX_SET)
	 { mutex_destroy (&info->lowMux);
	 }
	if (info->resources & PMC_INTR_MUTEX_SET)
	 { mutex_destroy (&info->intrMux);
	 }
#if 0
	if (info->resources & PMC_REGS_MAPPED)
	 { ddi_regs_map_free (&info->regsHandle);
	 }
#endif
	ddi_soft_state_free (pmci, info->instance);
}

static int pmcattach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	PmcSoftc *info;
	int instance;

	static char *mainNames[] = { "0", "1", "2", "3" };
	static char *chanNames[] = { "ChanX_0", "ChanX_1", 
				     "ChanX_2", "ChanX_3" };
	int k;

	switch (cmd)
	 { case DDI_ATTACH:
	    { instance = ddi_get_instance(dip);
	      if (instance >= MAX_INSTANCES)
	       { return (DDI_FAILURE);
	       }
	      if (ddi_soft_state_zalloc(pmci, instance) != 0)
	       { return (DDI_FAILURE);
	       }
	      info = ddi_get_soft_state(pmci, instance);

	      info->dip = dip;
	      info->instance = instance;
	      enioInit (&info->enio);	
#if 0
	      dev_attr.devacc_attr_version = DDI_DEVICE_ATTR_V0;
	      dev_attr.devacc_attr_endian_flags = DDI_NEVERSWAP_ACC;
	      dev_attr.devacc_attr_dataorder = DDI_STRICTORDER_ACC;
	      if (ddi_regs_map_setup (dip, PMC_REG_SPEC, 
		  (caddr_t*)&info->regs, /*offset=*/0, /*len=*/0,
		   &dev_attr, &info->regsHandle) == DDI_SUCCESS)
	       { cmn_err (CE_CONT, "pmc: regs_map_setup worked: 0x%x\n", 
			  info->regs);
		 info->resources |= PMC_REGS_MAPPED;
	       }
	      else
	       { cmn_err (CE_CONT, "pmc: regs_map_setup failed\n");
		 info->regs = 0;
		 pmcfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
#endif
	      info->iobase = 0x380;	/* XXX XXX XXX */

	      /* make device quiescent */
	      if (pmcHardwareReset (info) <0)
	       { cmn_err (CE_CONT, "pmc: hardware reset failed\n");
		 pmcfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }

	      /* set up interrupt-level mutex for the device. Start
		 by assigning a dummy interrupt handler so the
		 mutex is associated with the proper interrupt context,
		 without risk of an a handler getting called. */

	      if (ddi_add_intr (dip, PMC_INTR_SPEC,&info->iblkCookie,
		     NULL, PMC_NULL_HANDLER, NULL) != DDI_SUCCESS)
	       { pmcfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }

	      mutex_init (&info->intrMux, "pmc intr mutex", MUTEX_DRIVER,
			  (void*)info->iblkCookie);
	      ddi_remove_intr (dip, PMC_INTR_SPEC, info->iblkCookie);
	      info->resources |= PMC_INTR_MUTEX_SET;

	      /* see if the device interrupt is delivered in a high-level
		 interrupt context. If it is, we need to set up a 
		 low-level interrupt handler and mutex as well,
		 because we can't call cmn_err() or pollwakeup()
		 from a high-level interrupt context. */

	      if (ddi_intr_hilevel (dip, PMC_INTR_SPEC))
	       { info->hasHighLevelIntr = 1;
		 
		 /* set up low-level mutex and handler */

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			PMC_NULL_HANDLER, NULL) != DDI_SUCCESS)
		  { pmcfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 mutex_init (&info->lowMux, "pmc low mutex", MUTEX_DRIVER,
			     (void*)info->softIblkCookie);
		 ddi_remove_softintr (info->softIntrId);
		 info->resources |= PMC_LOW_MUTEX_SET;

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			pmcsoftintr, (caddr_t)info) != DDI_SUCCESS)
		  { pmcfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 info->resources |= PMC_SOFT_IHANDLER_SET;

		 info->muxp = &info->lowMux;
		 cmn_err (CE_CONT, "pmc: interrupts are high level\n");
	       }
	      else
	       { info->hasHighLevelIntr = 0;
		 info->muxp = &info->intrMux;
		 cmn_err (CE_CONT, "pmc: interrupts are low level\n");
	       }

	      if (ddi_add_intr (dip, PMC_INTR_SPEC, &info->iblkCookie,
		     &info->idevCookie, pmcintr,(caddr_t)info) != DDI_SUCCESS)
	       { pmcfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= PMC_IHANDLER_SET;

	      if (ddi_create_minor_node (dip, mainNames[instance], S_IFCHR, 
		     instance, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { pmcfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= PMC_MINOR_CREATED;
	      for (k=0; k<PMC_NUM_CHANS; k++)
	       { chanNames[instance][4] = k+'0';
		 cmn_err (CE_CONT, "pmc: minor %d\n", k);
		 if (ddi_create_minor_node (dip, chanNames[instance], S_IFCHR, 
			instance|CHAN_MINOR(k), DDI_PSEUDO, 0) != DDI_SUCCESS)
		  { pmcfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
	       }
	      info->intrCnt = 0;
	// XXX init other info stuff here
	      ddi_report_dev(dip);
	      cmn_err (CE_CONT, "succeeded\n");
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }
}

static int pmcdetach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	PmcSoftc *info;
	
	switch (cmd)
	 { case DDI_DETACH:
	    { info = ddi_get_soft_state (pmci, ddi_get_instance(dip));

cmn_err (CE_CONT, "pmc: detach called\n");
	      /* make device quiescent */
	      pmcHardwareReset (info);
	      pmcfreeResources (dip, info);
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }	
}

static int pmcprobe (dev_info_t *dip)
{
        int iobase;
	int status = DDI_PROBE_SUCCESS;

	iobase = ddi_prop_get_int (DDI_DEV_T_ANY, dip, 
		     DDI_PROP_DONTPASS|DDI_PROP_NOTPROM, "ioaddr", -1);
	if (iobase == -1)
	 { cmn_err (CE_CONT, "pmc probe: can't get property 'ioaddr'\n");
	   return (DDI_PROBE_FAILURE);
	 }
	if (!pmcHardwareProbe (iobase))
	 { status = DDI_PROBE_FAILURE;
	 }
	return (status);
}

static u_int pmcintr (caddr_t arg)
{
	/* Called in response to an interrupt from the PMC */

	PmcSoftc *info = (PmcSoftc*)arg;
	int needSoftIntr = 0;
	int pending;
	int stray;

	mutex_enter (&info->intrMux);
	pending = pmcHardwareIntrHandle (info, &stray);
	if (pending == 0 && stray == 0)
	 { mutex_exit (&info->intrMux);
	   return (DDI_INTR_UNCLAIMED);
	 }

	if (pending)
	 { if (info->hasHighLevelIntr)
	    { if (!info->softIntrRunning)
	       { needSoftIntr = 1;
	       }
	    }
	   else
	    { 
	      info->intrCnt++;
	      // handle interrupt here
	    }
	 }
	if (stray)
	 { /* declare a stray interrupt */
	   if (info->hasHighLevelIntr)
	    { info->stray = stray;
	      if (!info->softIntrRunning)
	       { needSoftIntr = 1;
	       }
	    }
	   else
	    { cmn_err (CE_CONT, "pmc%d: stray interrupt(s); mask=0x%x\n",
		   		 info->instance, stray);
	    }
	 }
	mutex_exit (&info->intrMux);
	if (needSoftIntr)
	 { ddi_trigger_softintr (info->softIntrId);
	 }
	return (DDI_INTR_CLAIMED);
}

static u_int pmcsoftintr (caddr_t arg)
{
	/* Used only when the device interrupts occur at high interrupt
	   priority. Calls pollwakeup and cmn-err because the high-level
	   interrupt handler can't. */

	PmcSoftc *info = (PmcSoftc*)arg;
	int stray;

	mutex_enter (&info->lowMux);
	mutex_enter (&info->intrMux);

	/* don't know why info->softIntrRunning needs to be checked
	   in the next phrase, but it was in the manual example */
	if (info->stray == 0 || info->softIntrRunning)
	 { mutex_exit (&info->intrMux);
	   mutex_exit (&info->lowMux);
	   return (DDI_INTR_UNCLAIMED);	
	 }
	info->softIntrRunning = 1;

	if ((stray = info->stray) != 0)
	 { info->stray = 0;
	 }
	mutex_exit (&info->intrMux);

	if (stray)
	 { cmn_err (CE_CONT, "pmc%d: stray interrupt(s); mask=0x%x\n",
			      info->instance, stray);
	 }
	info->intrCnt++;

	// handle interrupt here

	info->softIntrRunning = 0;
	mutex_exit (&info->lowMux);
	return (DDI_INTR_CLAIMED);	
}

STPMCC int pmcopen (dev_t *dev, int flag, int otyp, cred_t *crp)
{
	/* This device is an exclusive open device: only one process
	   may open it at a time. This prevents contention problems
	   and makes the driver software easier to implement. */

	PmcSoftc *info;
	int chan;

	if ((info = ddi_get_soft_state(pmci,GET_INSTANCE(*dev))) == NULL)
	 { return (ENXIO);
	 }
	if (otyp != OTYP_CHR)
	 { return (EINVAL);
	 }
	chan = GET_CHAN_DEV(*dev);
	mutex_enter (info->muxp);
/* enforce exclusive open on channel devices */
	if (chan >= 0)
	 { if (info->busy[chan])
	    { mutex_exit (info->muxp);
	      return EBUSY;
	    }
	   else
	    { info->busy[chan] = 1;
	    }
	 }
	if (!(info->flags & PMC_OPEN))
	 { 
	   if (pmcHardwareOpen (info) <0)
	    { cmn_err (CE_CONT, "pmc: hardware open failed\n");
	      return (ENXIO);
	    }
	   info->flags |= PMC_OPEN;
	 }
	mutex_exit (info->muxp);
	return (0);
}

STPMCC int pmcclose (dev_t dev, int flag, int otyp, cred_t *crp)
{
	PmcSoftc *info;
	int chan;

	chan = GET_CHAN_DEV(dev);
	if ((info = ddi_get_soft_state(pmci,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	cmn_err (CE_CONT, "close chan %d\n", chan);
	mutex_enter (info->muxp);
	if (pmcHardwareClose (info) < 0)
	 { cmn_err (CE_CONT, "pmc: hardware close failed\n");
	 }
	info->flags = 0;
	if (chan != -1)
	 { info->busy[chan] = 0;
	 }
	mutex_exit (info->muxp);
	return (0);
}

STPMCC int pmcread (dev_t dev, uio_t *uiop, cred_t *crp)
{
	/* The read routine returns whatever force/torque data 
	   may have been queued */

	return (ENXIO);
}

STPMCC int pmcwrite (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STPMCC int pmcmmap (dev_t dev, off_t off, int prot)
{
	return (-1);
}

STPMCC int pmcchpoll (dev_t dev, short events, int anyyet,
		      short *reventsp, struct pollhead **phpp)
{
	return (ENXIO);
}

STPMCC int pmcioctl (dev_t dev, int cmd, int arg, int mode, 
		     cred_t *crp, int *rvalp)
{
	PmcSoftc *info;

	if ((info = ddi_get_soft_state(pmci,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case PMCIOC_GET_ICNT:
	    { 
	      if (copyout((caddr_t)&info->intrCnt, (caddr_t)arg,sizeof(int)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case PMCIOC_SET_PIT:
	    { if (arg == 0)
	       { pmcHardwareIntrEnable (info, 0);
		 pmcHardwareSetPit (info, 0);
	       }
	      else
	       { pmcHardwareSetPit (info, arg);
		 pmcHardwareIntrEnable (info, 1);
	       }
	      break;
	    }
	   case PMCIOC_ENABLE_IOPORTS:
	    {
	      if (arg)
	       { enioCall0 (&info->enio);
	       }
	      else
	       { enioCall1 (&info->enio);
	       }
	      break;
	    }
	 }
	return (0);
}

/*------------------------ */

#define READW(iob,a)	(outw(iob+PMC_CTRL,a), inw(iob+PMC_XFER))
#define WRITEW(iob,a,w)	(outw(iob+PMC_CTRL,a), outw(iob+PMC_XFER,w))

static int pmcHardwareProbe (int base)
{
	unsigned long regval;

	pmcWriteDSPReg (base, 0x71, 2);
	regval = pmcReadDSPReg (base, 0x71);
	if (regval != 2 && regval != 1 && regval != 0)
	 { return 0;
	 }
	else
	 { return 1;
	 }
}

static int pmcHardwareReset (PmcSoftc *info)
{
	/* make the device quiescent. Called at attach and detach time */

	WRITEW (info->iobase, PMC_DSPRESET, 1);
	WRITEW (info->iobase, PMC_DSPRESET, 0);
	return 0;
}

static int pmcHardwareClose(PmcSoftc *info)
{
	/* make the device quiet when it is closed */

	/* disable interrupts */
	return (pmcHardwareIntrEnable (info, 0));
	return 0;
}

static int pmcHardwareOpen (PmcSoftc *info)
{
	/* initialize the device when it is openned */	
#if 0
	if (pmcReadCom (PMC_CARD_DATA_CMD, PMC_READ, &info->cdata,
			sizeof(info->cdata)/2, info) < 0)
	 { return (-1);
	 }
#endif
	return (0);
}

static int pmcHardwareIntrEnable (PmcSoftc *info, int enable)
{
	/* enable/disable device interrupts */

	if (enable)
	 { WRITEW (info->iobase, PMC_INTENAHOST, PMC_PIT_INTENT);
cmn_err (CE_CONT, "enable\n");
	 }
	else
	 { WRITEW (info->iobase, PMC_INTENAHOST, 0);
cmn_err (CE_CONT, "disable\n");
	 }
	return (1);
}

static int pmcHardwareIntrHandle (PmcSoftc *info, int *stray)
{
	/* clear interrupts, return true if a valid interrupt is pending,
	   and return a mask of any stray interrupts in *stray. */

	unsigned short savedCtrl;
	unsigned short flags;
	int pending = 0;

	savedCtrl = inw(info->iobase+PMC_CTRL);
	flags = READW(info->iobase,PMC_INTFLGHOST);
	*stray = 0;
	if (flags)
	 { WRITEW (info->iobase, PMC_INTFLGHOST, flags);
	   if ((info->flags & PMC_OPEN) && (flags & PMC_PIT_INTENT))
	    { pending = 1;
	    }
	   if (!(info->flags & PMC_OPEN) || (flags & ~PMC_PIT_INTENT))
	    { *stray = 1;
	    }
	 }
	outw(info->iobase+PMC_CTRL, savedCtrl);
	return pending;
}

static void pmcHardwareSetPit (PmcSoftc *info, int arg)
{
	if (arg)
	 { WRITEW (info->iobase, PMC_PIT, 0xffff & arg);
	   WRITEW (info->iobase, PMC_PITH, 0xffff & (arg >> 16));
	   WRITEW (info->iobase, PMC_GATE, 1);
cmn_err (CE_CONT, "wrote %x %x\n", 0xffff & arg, 0xffff & (arg >> 16));
	 }
	else
	 { 
	   WRITEW (info->iobase, PMC_GATE, 0);
cmn_err (CE_CONT, "gate 0, cnt=%d\n", info->intrCnt);
	 }
}

static unsigned short pmcReadReg (int iobase, int reg)
{
	return READW (iobase, reg);
}

static void pmcReadRegs (int iobase, unsigned short *buf, int n, int reg)
{
	int i;
	outw (iobase+PMC_CTRL, reg);
	for (i=0; i<n; i++)
	 { buf[i] = inw(iobase+PMC_XFER);
	 }	
}

static unsigned long pmcReadDSPReg (int iobase, int reg)
{
	unsigned short buf[2];
	outw (iobase+PMC_CTRL, PMC_DSPREG|reg);
	buf[0] = inw(iobase+PMC_XFER);
	buf[1] = inw(iobase+PMC_XFER);
	return (buf[0] + (buf[1] << 16));
}

static void pmcWriteReg (int iobase, int reg, unsigned short val)
{
	WRITEW (iobase, reg, val);
}

static void pmcWriteRegs (int iobase, int reg, unsigned short *buf, int n)
{
	int i;
	outw (iobase+PMC_CTRL, reg);
	for (i=0; i<n; i++)
	 { outw (iobase+PMC_XFER, buf[i]);
	 }
}

static void pmcWriteDSPReg (int iobase, int reg, unsigned long val)
{
	outw (iobase+PMC_CTRL, PMC_DSPREG|reg);
	outw (iobase+PMC_XFER, val & 0xffff);
	outw (iobase+PMC_XFER, val >> 16);
}
